"""Concrete implementation with nidaqmx-python (:mod:`pymanip.aiodaq.daqmx`)
============================================================================

This module implements a concrete implementation of the
:class:`~pymanip.aiodaq.AcquisitionCard` class using the :mod:`nidaqmx`
module.

.. autoclass:: DAQmxSystem
   :members:
   :private-members:

.. autofunction:: get_device_list

"""

import asyncio
import time

import numpy as np

from nidaqmx import Task
from nidaqmx.constants import READ_ALL_AVAILABLE
from nidaqmx.errors import DaqError
from nidaqmx.constants import TerminalConfiguration
from nidaqmx.system import system, device

from pymanip.aiodaq import (
    TerminalConfig,
    TriggerConfig,
    AcquisitionCard,
    TimeoutException,
)

ConcreteTerminalConfig = {
    TerminalConfig.RSE: TerminalConfiguration.RSE,
    TerminalConfig.NRSE: TerminalConfiguration.NRSE,
    TerminalConfig.Diff: TerminalConfiguration.DIFFERENTIAL,
    TerminalConfig.PseudoDiff: TerminalConfiguration.PSEUDODIFFERENTIAL,
}


class DAQmxSystem(AcquisitionCard):
    """This class is the concrete implementation for NI DAQmx board using
    the :mod:`nidaqmx` module.
    """

    def __init__(self):
        """Constructor method
        """
        super(DAQmxSystem, self).__init__()
        self.task = Task()
        self.reading = False
        self.stop_lock = asyncio.Lock()
        self.read_lock = asyncio.Lock()

    @property
    def samp_clk_max_rate(self):
        """Maximum sample clock rate
        """
        return self.task.timing.samp_clk_max_rate

    def possible_trigger_channels(self):
        """This method returns the list of channels that can be used as trigger.
        """
        return [chan.name for chan in self.channels]

    def close(self):
        """This method closes the active task, if there is one.
        """
        if self.task:
            self.channels = []
            self.task.close()
            self.task = None

    def add_channel(self, channel_name, terminal_config, voltage_range):
        """Concrete implementation of :meth:`pymanip.aiodaq.AcquisitionCard.add_channel`.

        .. todo::
            Actually check the type for terminal_config.

        """
        tc = ConcreteTerminalConfig[terminal_config]
        ai_chan = self.task.ai_channels.add_ai_voltage_chan(
            channel_name,
            terminal_config=tc,
            min_val=-voltage_range,
            max_val=voltage_range,
        )
        self.channels.append(ai_chan)
        self.actual_ranges.append(ai_chan.ai_max)

    def configure_clock(self, sample_rate, samples_per_chan):
        """Concrete implementation of :meth:`pymanip.aiodaq.AcquisitionCard.configure_clock`
        """
        self.task.timing.cfg_samp_clk_timing(
            sample_rate, samps_per_chan=samples_per_chan
        )
        self.sample_rate = sample_rate
        self.samples_per_chan = samples_per_chan

    def configure_trigger(
        self,
        trigger_source=None,
        trigger_level=0,
        trigger_config=TriggerConfig.EdgeRising,
    ):
        """Concrete implementation of :meth:`pymanip.aiodaq.AcquisitionCard.configure_trigger`

        .. todo::
           implement trigger_config other than the defaults value

        """

        st = self.task.triggers.start_trigger
        if trigger_source is None:
            print("disable_start_trig")
            st.disable_start_trig()
        else:
            if trigger_config != TriggerConfig.EdgeRising:
                raise NotImplementedError()  # TODO
            print(f"cfg_anlg_edge_start_trig({trigger_source:}, {trigger_level:})")
            st.cfg_anlg_edge_start_trig(trigger_source, trigger_level=trigger_level)

    def start(self):
        """This method starts the task.
        """
        self.task.start()
        self.running = True

    async def stop(self):
        """This asynchronous method aborts the current task.
        """
        async with self.stop_lock:
            if self.running:
                self.running = False
                while self.reading:
                    await asyncio.sleep(1.0)
                self.task.stop()

    async def read(self, tmo=None):
        """This asynchronous method reads data from the task.
        """
        async with self.read_lock:
            self.reading = True
            done = False
            start = time.monotonic()
            while self.running:
                try:
                    await self.loop.run_in_executor(
                        None, self.task.wait_until_done, 1.0
                    )
                except DaqError:
                    if tmo and time.monotonic() - start > tmo:
                        raise TimeoutException()
                    else:
                        continue
                done = True
                break
            if done and self.running:
                data = await self.loop.run_in_executor(
                    None, self.task.read, READ_ALL_AVAILABLE
                )
                self.last_read = time.monotonic()
                data = np.array(data)
            else:
                data = None
            self.reading = False
            return data


def get_device_list():
    """This function returns the list of devices that the NI DAQmx library
    can discover.

    :return: dictionnary with board description as key and channels as value
    :rtype: dict

    """
    sys = system.System()
    device_list = dict()
    for devname in sys.devices.device_names:
        dev = device.Device(devname)
        description = dev.product_type
        if description.startswith("PXI"):
            description = (
                f"PXI {dev.pxi_chassis_num:d} "
                f"Slot {dev.pxi_slot_num:d} "
                f"({dev.product_type:})"
            )
        elif description.startswith("PCI"):
            description = (
                f"{dev.product_type:} " f"({dev.pci_bus_num:} " f"{dev.pci_dev_num:})"
            )
        device_list[description] = dev.ai_physical_chans.channel_names
    return device_list



#! /usr/bin/env python
"""

This module contains miscellaneous GPIB functions:

 - scanGpib

"""

from __future__ import print_function, unicode_literals

import platform
import warnings

if platform.system() == "Linux":
    try:
        import gpib

        has_gpib = True
    except Exception:
        print("Warning: Linux-GPIB Python bindings not found")
        has_gpib = False
else:
    has_gpib = False


def scanGpib(board):
    if not has_gpib:
        warnings.warn(
            "Linux-GPIB is not available. ScanGpib with VISA not implemented at the moment.",
            RuntimeWarning,
            stacklevel=2,
        )
    else:
        for pad in range(1, 31):
            listen = gpib.listener(board, pad)
            # print(board, pad, listen)
            if listen:
                print("GPIB" + str(board) + "::" + str(pad))
                try:
                    ud = gpib.dev(board, pad, 0, 10, 1, 0)
                    if ud > 0:
                        gpib.write(ud, "*CLS;*IDN?")
                        description = gpib.read(ud, 256)
                        print(description.strip().decode("ascii"))
                except Exception:
                    pass



"""PCO Camera module (:mod:`pymanip.video.pco.camera`)
======================================================

This module implement the :class:`pymanip.video.pco.PCO_Camera` class using
bindings to the Pixelfly library from :mod:`pymanip.video.pco.pixelfly`.

.. autoclass:: PCO_Camera
   :members:
   :private-members:

.. autofunction:: PCO_get_binary_timestamp

.. autoclass:: PCO_Buffer
   :members:
   :private-members:

"""

import sys
import itertools
import ctypes
import datetime
import win32event

import numpy as np
import asyncio

from pymanip.video import MetadataArray, Camera, CameraTimeout
import pymanip.video.pco.pixelfly as pf


def PCO_get_binary_timestamp(image):
    """This functions reads the BCD coded timestamp in the first 14 pixels of an image
    from a PCO camera.

    :param image: the PCO camera image buffer
    :type image: array
    :return: counter, timestamp
    :rtype: int, datetime

    We assume the following format (per PCO documentation):

    ========= =================== ======
    Pixel     Description         Range
    ========= =================== ======
    Pixel  1  Image counter (MSB) 00..99
    Pixel  2  Image counter       00..99
    Pixel  3  Image counter       00..99
    Pixel  4  Image counter (LSB) 00..99
    Pixel  5  Year (MSB)          20
    Pixel  6  Year (LSB)          03..99
    Pixel  7  Month               01..12
    Pixel  8  Day                 01..31
    Pixel  9  Hour                00..23
    Pixel 10  Minutes             00..59
    Pixel 11  Seconds             00..59
    Pixel 12  µs * 10000          00..99
    Pixel 13  µs * 100            00..99
    Pixel 14  µs                  00..90
    ========= =================== ======

    """
    counter = pf.bcd_to_int(image[:4], endianess="big")
    year = pf.bcd_to_int(image[4:6], endianess="big")
    month = pf.bcd_to_int(image[6])
    day = pf.bcd_to_int(image[7])
    hour = pf.bcd_to_int(image[8])
    minutes = pf.bcd_to_int(image[9])
    seconds = pf.bcd_to_int(image[10])
    microseconds = pf.bcd_to_int(image[11:], endianess="big")
    return (
        counter,
        datetime.datetime(year, month, day, hour, minutes, seconds, microseconds),
    )


class PCO_Buffer:
    """This class represents an allocated buffer for the PCO camera.
    It implements context manager, as well as utility function to convert to bytes
    and numpy array. The buffer is allocated in the constructor method, and freed
    either by the context manager exit method, or manually calling the :meth:`free`
    method.

    :param cam_handle: camera handle
    :type cam_handle: HANDLE
    :param XResAct: resolution in x direction
    :type XResAct: int
    :param YResAct: resolution in y direction
    :type YResAct: int
    """

    def __init__(self, cam_handle, XResAct, YResAct):
        """Constructor method
        """
        self.cam_handle = cam_handle
        self.XResAct = XResAct
        self.YResAct = YResAct
        bufSizeInBytes = XResAct * YResAct * ctypes.sizeof(ctypes.wintypes.WORD)

        self.bufPtr = ctypes.POINTER(ctypes.wintypes.WORD)()
        num, event = pf.PCO_AllocateBuffer(cam_handle, -1, bufSizeInBytes, self.bufPtr)
        self.bufNr = num
        self.event_handle = event

    def free(self):
        """This methods frees the buffer.
        """
        pf.PCO_FreeBuffer(self.cam_handle, self.bufNr)
        self.bufPtr = None

    def __enter__(self):
        """Context manager enter method
        """
        return self

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        self.free()

    def as_array(self):
        """This methods returns the buffer as a numpy array. No data is copied,
        the memory is still bound to this buffer. The user must copy the data if
        necessary.

        :return: image array
        :rtype: numpy.ndarray

        """
        return np.ctypeslib.as_array(self.bufPtr, shape=(self.YResAct, self.XResAct))

    def bytes(self):
        """This methods returns the data as a bytearray.

        :return: image data
        :rtype: bytearray

        """
        nval = self.XResAct * self.YResAct
        bufType = ctypes.wintypes.WORD * nval
        return bytearray(bufType.from_address(ctypes.addressof(self.bufPtr.contents)))


class PCO_Camera(Camera):
    """Concrete :class:`pymanip.video.Camera` class for PCO camera.

    :param interface: interface where to look for the camera, defaults to "all"
    :type interface: str, optional
    :param camera_num: camera number to look for, defaults to 0.
    :type camera_num: int, optional
    :param metadata_mode: enable PCO Metadata mode, defaults to False.
    :type metadata_mode: bool, optional
    :param timestamp_mode: enable Timestamp mode (supported by all cameras), defaults to True.
    :type timestamp_mode: bool, optional

    """

    def __init__(
        self, interface="all", camera_num=0, *, metadata_mode=False, timestamp_mode=True
    ):
        """Constructor method

        .. note::
            pco.sdk_manual page 10:
            First step is to PCO_OpenCamera
            As next step camera description and status should be queried
            by calling PCO_GetCameraDescription and PCO_GetCameraHealthStatus
        """

        print("interface =", interface)
        print("camera_num =", camera_num)

        self.handle = pf.PCO_OpenCameraEx(interface, camera_num)
        self.camera_description = pf.PCO_GetCameraDescription(self.handle)
        warn, err, status = self.health_status()
        if warn or err:
            print("Warning bits :", warn)
            print("Error bits :", err)
        else:
            print("Connected to", pf.PCO_GetInfoString(self.handle))
            print("Status bits :", status)
        pf.PCO_SetBitAlignment(self.handle, sys.byteorder == "little")
        self.metadata_mode = metadata_mode
        self.timestamp_mode = timestamp_mode
        if timestamp_mode:
            # Timestamp is supported by all cameras but the information
            # is written on the first 14 pixels of the transfered image
            pf.PCO_SetTimestampMode(
                self.handle, 0x0001
            )  # binary mode (BCD coded in the first 14 px)
        else:
            pf.PCO_SetTimestampMode(self.handle, 0x0000)
        if metadata_mode:
            # MetaData is supported on pco.dimax and pco.edge only
            MetaDataSize, MetaDataVersion = pf.PCO_SetMetaDataMode(self.handle, True)
            self.MetaDataSize = MetaDataSize
            self.MetaDataVersion = MetaDataVersion

    def close(self):
        """This method closes the connection to the camera.
        """
        pf.PCO_CloseCamera(self.handle)
        self.handle = None
        print("Connection to camera closed.")

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        super(PCO_Camera, self).__exit__(type_, value, cb)
        self.close()

    # Query states
    def health_status(self):
        """This method queries the camera for its health status.

        :return: warn, err, status
        """
        warn, err, status = pf.PCO_GetCameraHealthStatus(self.handle)
        return warn, err, status

    # Set camera settings
    def set_adc_operating_mode(self, mode):
        """This function selects single or dual ADC operating mode:

           :param mode: "single" (or 0x0001) or "dual" (or 0x0002)
           :type mode: int or str

           - Single mode increases linearity;

           - Dual mode allows higher frame rates.

        """
        if mode not in (0x0001, 0x0002):
            shortcut = {"single": 0x0001, "dual": 0x0002}
            mode = shortcut[mode]
        pf.PCO_SetADCOperation(self.handle, mode)

    def set_pixel_rate(self, rate):
        """This function selects the pixel rate for sensor readout.

        :param rate: readout rate (in Hz)
        :type rate: float

        For PCO.1600: 10 Mhz or 40 MHz
        """
        pf.PCO_SetPixelRate(self.handle, int(rate))

    def set_trigger_mode(self, mode):
        """This method sets the trigger mode for the camera.

        :param mode: one of PCO_TriggerModeDescription
        :type mode: str

        Possible values are:

        ======= ==============================================
        mode    description
        ======= ==============================================
        0x0000  auto sequence
        0x0001  software trigger
        0x0002  external exposure start & software trigger
        0x0003  external exposure control
        0x0004  external synchronized
        0x0005  fast external exposure control
        0x0006  external CDS control
        0x0007  slow external exposure control
        0x0102  external synchronized HDSDI
        ======= ==============================================

        """

        if isinstance(mode, bool):
            if mode:
                self.set_trigger_mode(0x0002)
            else:
                self.set_trigger_mode(0x0000)
        elif mode in pf.PCO_TriggerModeDescription:
            pf.PCO_SetTriggerMode(self.handle, mode)
        else:
            for key, val in pf.PCO_TriggerModeDescription.items():
                if val == mode:
                    break
            else:
                raise ValueError("Unknown trigger mode : " + str(mode))
            pf.PCO_SetTriggerMode(self.handle, key)

    def set_delay_exposuretime(self, delay=None, exposuretime=None):
        """This method sets both the delay and the exposure time.

        :param delay: delay in seconds
        :type delay: float
        :param exposuretime: exposure time in seconds
        :type exposuretime: float
        """
        if delay is None or exposuretime is None:
            (
                delay_current,
                exposure_current,
                tb_delay,
                tb_exposure,
            ) = pf.PCO_GetDelayExposureTime(self.handle)
        if delay is None:
            delay = delay_current
        else:
            delay = delay * 1000
            tb_delay = 0x0002
            if delay < 1.0:
                delay = delay * 1000
                tb_delay = 0x0001
                if delay < 1.0:
                    delay = delay * 1000
                    tb_delay = 0x0000
        if exposuretime is None:
            exposuretime = exposure_current
        else:
            exposuretime = exposuretime * 1000
            tb_exposure = 0x0002
            if exposuretime < 1.0:
                exposuretime = exposuretime * 1000
                tb_exposure = 0x0001
                if exposuretime < 1.0:
                    exposuretime = exposuretime * 1000
                    tb_exposure = 0x0000
        units = {0x0000: "ns", 0x0001: "µs", 0x0002: "ms"}
        print("Setting delay to", int(delay), units[tb_delay])
        print("Setting exposure time to", int(exposuretime), units[tb_exposure])
        pf.PCO_SetDelayExposureTime(
            self.handle, int(delay), int(exposuretime), tb_delay, tb_exposure
        )

    def set_roi(self, roiX0=0, roiY0=0, roiX1=0, roiY1=0):
        r"""This method sets the positions of the upper left corner (X0,Y0) and lower right
        (X1,Y1) corner of the ROI (region of interest) in pixels.

        :param roiX0: left border in pixels, must be :math:`1 + 32n, n \in \mathbb{N}`
        :type roiX0: int
        :param roiY0: top border in pixels, must be :math:`1 + 8n, n \in \mathbb{N}`
        :type roiY0`: int
        :param roiX1: right border in pixels, must be :math:`32m, m \in \mathbb{N}`
        :type roiX1: int
        :param roiY1: bottom border in pixels, must be :math:`8m, m \in \mathbb{N}`

        The minimum ROI is :math:`64\times 16` pixels, and it is required that :math:`roiX1 \geq roix0` and :math:`roiY1 \geq roiY0`.
        """
        if (roiX0 - 1) % 32 != 0 or roiX1 % 32 != 0:
            raise ValueError("X0 must be 1+32n, X1 must be 32m, n, m integers")
        if (roiY0 - 1) % 8 != 0 or roiY1 % 8 != 0:
            raise ValueError("Y0 must be 1+8n, Y1 must be 8m, n, m integers")
        if roiX1 - roiX0 + 1 < 64 or roiY1 - roiY0 + 1 < 16:
            raise ValueError("Minimum ROI is 64 x 16 pixels")
        if roiX0 >= roiX1 or roiY0 >= roiY1:
            raise ValueError(
                "ROI expected xmin, ymin, xmax, ymax with xmax > xmin and ymax > ymin"
            )
        print(
            "Setting the ROI to (X0,Y0,X1,Y1)",
            int(roiX0),
            int(roiY0),
            int(roiX1),
            int(roiY1),
        )
        pf.PCO_SetROI(self.handle, int(roiX0), int(roiY0), int(roiX1), int(roiY1))

    def set_frame_rate(self, Frameratemode, Framerate, Framerateexposure):
        """This method sets Frame rate (mHz) and exposure time (ns).

        :param Frameratemode: one of the possible framerate modes (0x0000, 0x0001, 0x0002, 0x0003)
        :type Frameratemode: int
        :param Framerate: framerate in mHz
        :type Framerate: int
        :param Framerateexposure: Exposure time in ns
        :type Framerateexposure: int

        :return: message, framerate, exposure time
        :rtype: int, int, int

        The meaning of the framerate mode is given in this table

        ==============  =====================================================================
        Framerate mode  Meaning
        ==============  =====================================================================
        0x0000          Auto mode (camera decides which parameter will be trimmed)
        0x0001          Frame rate has priority, (exposure time will be trimmed)
        0x0002          Exposure time has priority, (frame rate will be trimmed)
        0x0003          Strict, function shall return with error if values are not possible.
        ==============  =====================================================================

        The message value in return gives the limiting factors when the condition are not fulfilled.
        The meaning is given in this table

        ==============  =========================================================================
        Message         Meaning
        ==============  =========================================================================
        0x0000          Settings consistent, all conditions met
        0x0001          Frame rate trimmed, frame rate was limited by readout time
        0x0002          Frame rate trimmed, frame rate was limited by exposure time
        0x0004          Exposure time trimmed, exposure time cut to frame time
        0x8000          Return values dwFrameRate and dwFrameRateExposure are not yet validated.
        ==============  =========================================================================

        In the case where message 0x8000 is returned, the other values returned are simply
        the parameter values passed to the function.
        """
        print(
            "Setting the frame rate and the exposure time to",
            Framerate,
            Framerateexposure,
        )
        message, f, te = pf.PCO_SetFrameRate(
            self.handle, Frameratemode, Framerate, Framerateexposure
        )
        return message, f, te

    def set_noise_filter_mode(self, mode):
        """This method does set the image correction operating mode of the camera. Image correction can either be switched
        totally off, noise filter only mode or noise filter plus hot pixel correction mode.
        The command will be rejected, if Recording State is [run], see PCO_GetRecordingState.

        :param mode: the noise filter mode
        :type mode: int

        The noise filter mode:

        ======= =============================
        Value   Mode
        ======= =============================
        0x0000  [off]
        0x0001  [on]
        0x0101  [on + hot pixel correction]
        ======= =============================

        """
        pf.PCO_SetNoiseFilterMode(self.handle, mode)

    def current_noise_filter_mode(self):
        """This methods queries the current noise filter mode.

        :return: the noise filter mode
        :rtype: int

        The noise filter mode:

        ======= =============================
        Value   Mode
        ======= =============================
        0x0000  [off]
        0x0001  [on]
        0x0101  [on + hot pixel correction]
        ======= =============================

        """
        return pf.PCO_GetNoiseFilterMode(self.handle)

    # Properties
    @property
    def resolution(self):
        """Camera maximum resolution
        """
        return self.camera_description.maximum_resolution_std

    @property
    def name(self):
        """Camera name
        """
        # if hasattr(self, '_name'):
        #    return self._name
        # PCO_GetCameraName is not supported by pco.1600
        # self._name = pf.PCO_GetCameraName(self.handle)
        # return self._name
        return "PCO Camera"

    @property
    def bitdepth(self):
        """Camera sensor bit depth
        """
        return 16

    def current_delay_exposure_time(self):
        """This method returns current delay and exposure time in seconds.

        :return: delay, exposure
        :rtype: float, float
        """

        delay, exposure, tb_delay, tb_exposure = pf.PCO_GetDelayExposureTime(
            self.handle
        )
        return (
            delay * pf.PCO_Timebases[tb_delay],
            exposure * pf.PCO_Timebases[tb_exposure],
        )

    def current_trigger_mode_description(self):
        """This method returns the current trigger mode description.

        :return: description of current trigger mode
        :rtype: str
        """
        return pf.PCO_TriggerModeDescription[pf.PCO_GetTriggerMode(self.handle)]

    def current_adc_operation(self):
        """This method returns the current ADC operation mode.

        :return: Current ADC operation mode (0x0001 for "single", 0x0002 for "dual")
        :rtype: int
        """
        return pf.PCO_GetADCOperation(self.handle)

    def current_pixel_rate(self):
        """This method returns the current pixel rate.

        :return: Current pixel rate (e.g. 10 MHz or 40 MHz for the PCO.1600
        :rtype: int
        """
        return pf.PCO_GetPixelRate(self.handle)

    @property
    def possible_pixel_rate(self):
        return [r for r in self.camera_description.dwPixelRateDESC]

    def current_frame_rate(self):
        """This method returns the current frame rate.

        :return: Current frame rate
        :rtype: int
        """
        return pf.PCO_GetFrameRate(self.handle)

    def acquisition_oneshot(self):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_oneshot` for the PCO camera.
        """
        # Arm camera
        pf.PCO_ArmCamera(self.handle)
        XResAct, YResAct, XResMax, YResMax = pf.PCO_GetSizes(self.handle)

        with PCO_Buffer(self.handle, XResAct, YResAct) as buffer:
            try:
                pf.PCO_SetImageParameters(
                    self.handle,
                    XResAct,
                    YResAct,
                    pf.IMAGEPARAMETERS_READ_WHILE_RECORDING,
                )
                pf.PCO_SetRecordingState(self.handle, True)
                pf.PCO_GetImageEx(
                    self.handle, 1, 0, 0, buffer.bufNr, XResAct, YResAct, 16
                )
                array = buffer.as_array().copy()
            finally:
                pf.PCO_SetRecordingState(self.handle, False)
                pf.PCO_CancelImages(self.handle)
        return array

    async def acquisition_async(
        self,
        num=np.inf,
        timeout=None,
        raw=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_async` for the PCO camera.
        """

        loop = asyncio.get_event_loop()

        if timeout is None:
            delay, exposure = self.current_delay_exposure_time()
            timeout = int(max((2000 * exposure, 1000)))

        # Arm camera
        if pf.PCO_GetRecordingState(self.handle):
            pf.PCO_SetRecordingState(self.handle, False)
        pf.PCO_ArmCamera(self.handle)
        warn, err, status = self.health_status()
        if err != 0:
            raise RuntimeError("Camera has error status!")
        XResAct, YResAct, XResMax, YResMax = pf.PCO_GetSizes(self.handle)

        with PCO_Buffer(self.handle, XResAct, YResAct) as buf1, PCO_Buffer(
            self.handle, XResAct, YResAct
        ) as buf2, PCO_Buffer(self.handle, XResAct, YResAct) as buf3, PCO_Buffer(
            self.handle, XResAct, YResAct
        ) as buf4:

            buffers = (buf1, buf2, buf3, buf4)
            try:
                pf.PCO_SetImageParameters(
                    self.handle,
                    XResAct,
                    YResAct,
                    pf.IMAGEPARAMETERS_READ_WHILE_RECORDING,
                )
                pf.PCO_SetRecordingState(self.handle, True)
                for buffer in buffers:
                    pf.PCO_AddBufferEx(
                        self.handle, 0, 0, buffer.bufNr, XResAct, YResAct, 16
                    )
                count = 0
                buffer_ring = itertools.cycle(buffers)
                while count < num:
                    if (
                        count == 0
                        and initialising_cams is not None
                        and self in initialising_cams
                    ):
                        initialising_cams.remove(self)

                    waitstat = await loop.run_in_executor(
                        None,
                        win32event.WaitForMultipleObjects,
                        [buffer.event_handle for buffer in buffers],
                        0,
                        timeout,
                    )
                    if waitstat == win32event.WAIT_TIMEOUT:
                        if raise_on_timeout:
                            raise CameraTimeout(f"Timeout ({timeout:})")
                        else:
                            stop_signal = yield None
                            if not stop_signal:
                                continue
                            else:
                                break
                    for ii, buffer in zip(range(4), buffer_ring):
                        waitstat = await loop.run_in_executor(
                            None, win32event.WaitForSingleObject, buffer.event_handle, 0
                        )
                        if waitstat == win32event.WAIT_OBJECT_0:
                            win32event.ResetEvent(buffer.event_handle)
                            statusDLL, statusDrv = pf.PCO_GetBufferStatus(
                                self.handle, buffer.bufNr
                            )
                            if statusDrv != 0:
                                raise RuntimeError(
                                    "buffer {:} error status {:}".format(
                                        buffer.bufNr, statusDrv
                                    )
                                )
                            if raw:
                                data = {"buffer": buffer.bytes()}
                                if self.timestamp_mode:
                                    counter, dt = PCO_get_binary_timestamp(
                                        buffer.bufPtr[:14]
                                    )
                                    data["counter"] = counter
                                    data["timestamp"] = dt
                                stop_signal = yield data
                            else:
                                if self.metadata_mode:
                                    metadata = pf.PCO_GetMetaData(
                                        self.handle, buffer.bufNr
                                    )
                                    stop_signal = yield MetadataArray(
                                        buffer.as_array(), metadata=metadata
                                    )
                                elif self.timestamp_mode:
                                    counter, dt = PCO_get_binary_timestamp(
                                        buffer.bufPtr[:14]
                                    )
                                    stop_signal = yield MetadataArray(
                                        buffer.as_array(),
                                        metadata={"counter": counter, "timestamp": dt},
                                    )
                                else:
                                    stop_signal = yield buffer.as_array()
                            count += 1
                            pf.PCO_AddBufferEx(
                                self.handle, 0, 0, buffer.bufNr, XResAct, YResAct, 16
                            )
                        else:
                            break
                        if stop_signal:
                            break
                    if stop_signal:
                        break
            finally:
                pf.PCO_SetRecordingState(self.handle, False)
                pf.PCO_CancelImages(self.handle)
        if stop_signal:
            yield True


if __name__ == "__main__":
    h = pf.PCO_OpenCameraEx("USB 3.0", 0)
    pf.PCO_GetInfoString(h)
    pf.PCO_CloseCamera(h)



"""PXI acquisition module (:mod:`pymanip.daq.PXI`)
==================================================

This module is an attempt to implement simultaneous DAQmx and Scope acquisition
on a PXI chassis.
It is not functionnal at the moment.

"""

import numpy as np

import PyDAQmx
import niScope


def read_analog(
    Scope_slot,
    Scope_channels,
    DAQmx_channels,
    volt_range,
    samples_per_chan,
    sample_rate,
):
    """
    Simultaneous NI Scope / NI DAQmx acquisition
    """

    # Configure the scope
    scope = niScope.Scope(Scope_slot)
    scope.ConfigureClock("VAL_PXI_CLOCK", "VAL_NO_SOURCE", "VAL_RTSI_0", True)
    scope.ConfigureHorizontalTiming(sampleRate=sample_rate, numPts=samples_per_chan)
    scope.ConfigureVertical(channelList=Scope_channels, voltageRange=volt_range)
    scope.ConfigureTrigger("Immediate")
    scope.ExportSignal(niScope.NISCOPE_VAL_REF_TRIGGER, "VAL_RTSI_1", "VAL_RTSI_1")
    sampling = scope.ActualSamplingRate
    if sampling != sample_rate:
        print("Warning: sample rate changed to", sampling)

    # Configure DAQmx channels
    task = PyDAQmx.Task()
    for chan in DAQmx_channels:
        task.CreateAIVoltageChan(
            chan,
            "",
            PyDAQmx.DAQmx_Val_Diff,
            -volt_range,
            volt_range,
            PyDAQmx.DAQmx_Val_Volts,
            None,
        )
        task.SetChanAttribute(chan, PyDAQmx.DAQmx_AI_Coupling, PyDAQmx.DAQmx_Val_DC)

    task.CfgDigEdgeStartTrig("RTSI1", PyDAQmx.DAQmx_Val_Rising)
    task.CfgSampClkTiming(
        "OnboardClock",
        sampling,
        PyDAQmx.DAQmx_Val_Rising,
        PyDAQmx.DAQmx_Val_FiniteSamps,
        samples_per_chan,
    )
    task.CfgInputBuffer(samples_per_chan)

    # Start DAQmx task and initiate Scope acquisition
    task.StartTask()
    scope.InitiateAcquisition()

    # Read scope data
    data_scope = scope.Fetch(Scope_channels)
    length = scope.ActualRecordLength
    print("NI-Scope: {:} samples read at {:} Hz".format(length, sampling))
    scope.close()

    # Read DAQmx data
    timeout = float(10 * samples_per_chan / sample_rate)
    buffer_size_in_samps = int(samples_per_chan * len(DAQmx_channels))
    data_daqmx = np.zeros((buffer_size_in_samps,), dtype=np.float64)
    samples_per_chan_read = PyDAQmx.int32()
    task.ReadAnalogF64(
        samples_per_chan,
        timeout,
        PyDAQmx.DAQmx_Val_GroupByChannel,
        data_daqmx,
        buffer_size_in_samps,
        PyDAQmx.byref(samples_per_chan_read),
        None,
    )
    print("DAQmx: %d samples read." % samples_per_chan_read.value)

    return (data_scope, data_daqmx)


if __name__ == "__main__":

    from pymanip.instruments import HP33120a

    source = HP33120a("GPIB0::10")
    f = source.freq.get()
    print("Source frequency is", f)

    n = 10000
    fe = 20 * f

    (data_scope, data_daqmx) = read_analog(
        "PXI1Slot3", "0", ("PXI1Slot4/ai2",), 5.0, n, fe
    )
    t = np.arange(n) / fe
    import matplotlib.pyplot as plt

    plt.figure()
    plt.plot(1000 * t, data_scope, "ro", label="Scope")
    plt.plot(1000 * t, data_daqmx, "bs", label="DAQ")
    plt.xlabel("t (ms)")
    plt.legend()
    plt.show()



"""Scope acquisition module (:mod:`pymanip.daq.Scope`)
======================================================

This module implements a :func:`read_analog` similar to that of the
:mod:`~pymanip.daq.DAQmx` module, but for Scope devices.
It uses the :mod:`niScope` module from National Instruments.

.. autofunction:: read_analog

"""

from niScope import Scope
from platform import platform
import time
import pymanip.mytime as MI
from collections import Iterable
from fluiddyn.util.terminal_colors import cprint


def print_horodateur(samples_per_chan, sample_rate):
    if platform().startswith("Windows"):
        dateformat = "%A %d %B %Y - %X (%z)"
    else:
        dateformat = "%A %e %B %Y - %H:%M:%S (UTC%z)"
    starttime = time.time()
    starttime_str = time.strftime(dateformat, time.localtime(starttime))
    endtime = starttime + samples_per_chan / sample_rate
    endtime_str = time.strftime(dateformat, time.localtime(endtime))
    print("Scope: Starting acquisition: " + starttime_str)
    print(
        "       Expected duration: %.2f min" % (samples_per_chan / (60.0 * sample_rate))
    )
    print("       Expected end time: " + endtime_str)


def read_analog(
    scope_name,
    channelList="0",
    volt_range=10.0,
    samples_per_chan=100,
    sample_rate=1000.0,
    coupling_type="DC",
):
    """This function reads signal from a digital oscillosope.

    :param scope_name: name of the NI-Scope device (e.g. 'Dev3')
    :type scope_channel: str
    :param channelList: comma-separated string of channel number (e.g. "0")
    :type channelList: str
    :param volt_range: voltage range
    :type volt_range: float
    :param samples_per_chan: number of samples to read per channel
    :type samples_per_chan: int
    :param sample_rate: for 5922 60e6/n avec n entre 4 et 1200
    :type sample_rate: float
    :param coupling_type: 'DC', 'AC', 'GND'
    :type coupling_type: str
    """

    # Make sure scalars have the correct type, as it will otherwise
    # fail to convert to the corresponding Vi types
    samples_per_chan = int(samples_per_chan)
    sample_rate = float(sample_rate)
    channelList = str(channelList)
    numChannels = len(channelList.split(","))
    if isinstance(volt_range, Iterable):
        volt_range = [float(v) for v in volt_range]
    else:
        volt_range = float(volt_range)
        if numChannels > 1:
            volt_range = [volt_range] * numChannels

    scope = Scope(scope_name)
    print("Scope:", scope_name)
    scope.ConfigureHorizontalTiming(sampleRate=sample_rate, numPts=samples_per_chan)
    scope.NumRecords = 1
    if numChannels == 1:
        scope.ConfigureVertical(channelList=channelList, voltageRange=volt_range)
    else:
        for chan, v in zip(channelList.split(","), volt_range):
            print("Scope: setting chan {:s} voltage range to {:}.".format(chan, v))
            scope.ConfigureVertical(channelList=chan, voltageRange=v)
            actualRange = scope.ActualVoltageRange(chan)
            if actualRange != volt_range:
                print(
                    "Scope: actual range for chan {:s} is {:}".format(chan, actualRange)
                )
    scope.ConfigureTrigger("Immediate")
    sampling = scope.ActualSamplingRate
    length = scope.ActualRecordLength
    print_horodateur(length, sampling)
    if sampling != sample_rate:
        cprint.red("Warning: sampling frequency changed to {:} Hz.".format(sampling))
    if length != samples_per_chan:
        cprint.red("Warning: record length changed to {:d} points.".format(length))
    if numChannels == 1:
        vRange = scope.ActualVoltageRange(channelList)
        if vRange != volt_range:
            cprint.red("Warning: actual voltage range is {:} V.".format(vRange))
    else:
        for chan, v in zip(channelList.split(","), volt_range):
            vv = scope.ActualVoltageRange(chan)
            if vv != v:
                cprint.red(
                    "Warning: actual range for channel {:s} is {:} V.".format(chan, vv)
                )
    scope.InitiateAcquisition()
    duration = samples_per_chan / sampling
    MI.sleep(duration)
    data = scope.Fetch(channelList, timeout=duration)
    (l, c) = data.shape
    print("Scope: {:d} samples read.".format(l))
    scope.close()

    return tuple(data[:, i] for i in range(numChannels))



"""PCO module (:mod:`pymanip.video.pco`)
========================================

This module is a shortcut for the :class:`pymanip.video.pco.camera.PCO_Camera`
class. It also defines utility functions for PCO camera.

.. autofunction:: PCO_read_binary_file

.. autofunction:: print_available_pco_cameras

"""

import numpy as np
from .camera import PCO_Camera
from . import pixelfly as pf


def PCO_read_binary_file(f):
    """This functions reads PCO binary image file.
    """
    return np.fromfile(f, dtype="<i2").reshape((1200, 1600))


def print_available_pco_cameras():
    """This functions queries the Pixelfly library for available cameras,
    and prints the result.
    """
    cams = list()
    while True:
        try:
            h = pf.PCO_OpenCamera()
        except pf.PCO_Error:
            break
        if h == 0:
            break
        print(pf.PCO_GetInfoString(h))
        cams.append(h)
    for h in cams:
        pf.PCO_CloseCamera(h)



"""Asynchronous extension of Lauda chillers
===========================================

"""

from serial.serialutil import SerialException
import asyncio
import fluidlab.instruments.chiller.lauda as fl_lauda
from pymanip.aioinstruments.aiodrivers import AsyncDriver
from pymanip.aioinstruments.aiofeatures import AsyncValue


class AsyncLauda(AsyncDriver, fl_lauda.Lauda):
    async def __aenter__(self):
        try:
            await super().__aenter__()
        except SerialException:
            # for some reason we randomly get
            # Cannot configure port, something went wrong. Original message: PermissionError(13, 'Un périphérique attaché au système ne fonctionne pas correctement.', None, 31)
            # It is often enough to just wait and try again
            print("Initial communication attempt failed on serial port.")
            print("Retrying in 5 seconds")
            await asyncio.sleep(5.0)
            await super().__aenter__()
        identification = await self.interface.aquery("TYPE\r")
        identification = identification.decode("ascii")
        if identification not in fl_lauda.Lauda.Models:
            if len(identification) > 0:
                raise fl_lauda.LaudaException("Unsupported model: " + identification)

            else:
                raise fl_lauda.LaudaException(
                    "Cannot communicate with Lauda on " + str(self.port)
                )

        else:
            self.rom = fl_lauda.Lauda.Models[identification]
            print("Identification: " + identification)
        return self


class AsyncLaudaValue(AsyncValue, fl_lauda.LaudaValue):
    async def aget(self):
        result = await super().aget()
        result = result.decode("ascii")
        if len(result) < 1:
            print("result =", result)
            raise fl_lauda.LaudaException("Erreur de communication")
        elif result.startswith("ERR"):
            raise fl_lauda.LaudaException("Erreur Lauda: " + result)
        else:
            return float(result)

    async def aset(self, value):
        command = self.command_set.format(value).encode("ascii")
        await self._interface.awrite(command)
        await asyncio.sleep(self.pause_instrument)
        confirmation = await self._interface.aread()
        confirmation = confirmation.decode("ascii")
        if confirmation != "OK":
            print(confirmation)
            raise fl_lauda.LaudaException("Erreur de communication")


class AsyncLaudaOnOffValue(AsyncLaudaValue, fl_lauda.LaudaOnOffValue):
    async def aget(self):
        possible_answers = {
            0: True,  # Stand-by false, Device ON
            1: False,  # Stand-by true, Device OFF
        }
        if self._driver.rom in fl_lauda.LaudaOnOffValue.Supported_ROM:
            resultat = await super().aget()
            resultat = int(resultat)
            if resultat not in possible_answers.keys():
                raise fl_lauda.LaudaException(f"Unknown answer: '{resultat:}'")
            return possible_answers[resultat]
        else:
            # Those cannot go to stand-by, and are always on
            return True

    async def aset(self, value):
        present_state = await self.aget()
        if bool(value) is bool(present_state):
            print("Chiller already on this state")
            return
        if value:
            command = "START\r".encode("ascii")
        else:
            command = "STOP\r".encode("ascii")
        await self._interface.awrite(command)
        await asyncio.sleep(self.pause_instrument)
        confirmation = await self._interface.aread()
        confirmation = confirmation.decode("ascii")
        if confirmation != "OK":
            print(confirmation)
            raise fl_lauda.LaudaException("Erreur de communication")


class AsyncLaudaStatValue(AsyncValue, fl_lauda.LaudaStatValue):
    async def aget(self):
        result = await super().aget()
        result = str(result)
        if len(result) < 3:
            raise fl_lauda.LaudaException("Erreur de communication")

        elif result.startswith("ERR"):
            raise fl_lauda.LaudaException("Erreur Lauda: " + result)

        else:
            print(result)
            return {
                "overheat": True if (result[0] == "1") else False,
                "lowlevel": True if (result[1] == "1") else False,
                "pumperr": True if (result[2] == "1") else False,
                "controllererror1": True if (result[3] == "1") else False,
                "controllererror2": True if (result[4] == "1") else False,
            }


afeatures = [
    AsyncLaudaValue(
        "setpoint", command_get="IN_SP_00\r", command_set="OUT_SP_00 {:.2f}\r"
    ),
    AsyncLaudaStatValue("stat", command_get="STAT\r"),
    AsyncLaudaValue("temperature", command_get="IN_PV_00\r"),
    AsyncLaudaValue("waterlevel", command_get="IN_PV_05\r"),
    AsyncLaudaOnOffValue(),
    AsyncLaudaValue(
        "status",
        command_get="STATUS\r",
    ),  # 0: OK, -1: Fault
]

AsyncLauda._build_class_with_features(afeatures)



"""

Async Agilent 34970a: extension of fluidlab Agilent34970a class with asynchroneous scan/get/set

"""

from datetime import datetime
import numpy as np
import fluidlab.instruments.multiplexer.agilent_34970a as fl_agilent_34970a
from pymanip.aioinstruments.aioiec60488 import AsyncIEC60488


class AsyncAgilent34970aValue(fl_agilent_34970a.Agilent34970aValue):
    def _build_driver_class(self, Driver):

        super()._build_driver_class(Driver)

        async def aget(self, chanList, samplesPerChan=1, sampleRate=None, verbose=None):
            if verbose is None:
                verbose = samplesPerChan > 1
            result = await self._driver.ascan(
                chanList, self.function_name, samplesPerChan, sampleRate, verbose
            )
            if len(result) == 1:
                result = result[0]
            return result

        self.aget = aget.__get__(self, self.__class__)

        async def aset(self, channel, value, warn=True):
            if self.name == "vdc":
                await self._driver.awrite_vdc(channel, value)
            else:
                raise ValueError("Specified value cannot be written")

        self.aset = aset.__get__(self, self.__class__)


class AsyncAgilent34970a(AsyncIEC60488, fl_agilent_34970a.Agilent34970a):
    async def ascan(
        self, channelList, functionName, samplesPerChan, sampleRate, verbose=True
    ):

        try:
            # Checks if channelList is iterable
            numChans = len([x for x in channelList])
        except Exception:
            # If not, convert to 1-tuple
            channelList = (channelList,)
            numChans = 1

        # Max number of points: 50000
        if samplesPerChan * numChans > 50000:
            raise ValueError("Maximum number of samples is 50000 on Agilent 34970A")

        if samplesPerChan > 1:
            timeInterval = 1.0 / sampleRate
            if timeInterval < 1e-3:
                raise ValueError("The timer resolution of the Agilent 34970A is 1 ms")

        # Check that channel numbers are right
        badChans = [x for x in channelList if ((x < 100) and (x >= 400))]
        if len(badChans) > 0:
            raise ValueError(
                "Channels must be specified in the form scc, where s is "
                "the slot number (100, 200, 300), and cc is the channel "
                "number. For example, channel 10 on the slot 300 is referred "
                "to as 310."
            )

        # Set measurement type to desired function on specified channels
        msg = 'SENS:FUNC "' + functionName + '",(@'
        for i in range(numChans):
            msg = msg + str(channelList[i])
            if i < (numChans - 1):
                msg = msg + ","
            else:
                msg = msg + ")"
        await self.interface.awrite(msg)

        # Set range on specified channels
        for chan in channelList:
            if str(chan) in self.Range:
                # set channel to manual range
                await self.interface.awrite(
                    "SENS:"
                    + functionName
                    + ":RANG "
                    + str(self.Range[str(chan)])
                    + ",(@"
                    + str(chan)
                    + ")"
                )
            elif functionName != "TEMP":
                # set channel to Auto Range
                await self.interface.awrite(
                    "SENS:" + functionName + ":RANG:AUTO ON,(@" + str(chan) + ")"
                )

        # Set NPLC for specified channels
        for chan in channelList:
            if str(chan) in self.NPLC:
                # set NPLC to specified value
                await self.interface.awrite(
                    "SENS:"
                    + functionName
                    + ":NPLC "
                    + str(self.NPLC[str(chan)])
                    + ",(@"
                    + str(chan)
                    + ")"
                )
                if samplesPerChan > 1:
                    # warn if wrong value (50Hz line hard coded here)
                    tMoy = self.NPLC[str(chan)] / 50.0
                    if tMoy > 1.0 / sampleRate:
                        print(
                            "Warning: averaging for {:.1f} ms, and sample time is {:.1f} ms".format(
                                1000.0 * tMoy, 1000.0 / sampleRate
                            )
                        )
            elif samplesPerChan > 1:
                print("Warning: NPLC should be specified for acquisitions")

        # Set TK Type for specified channels (if TK channel and TkType defined)
        if functionName == "TEMP":
            for chan in channelList:
                if str(chan) in self.TkType:
                    # set Tk type to specified value
                    await self.interface.awrite(
                        "SENS:TEMP:TRAN:TC:TYPE "
                        + str(self.TkType[str(chan)])
                        + ",(@"
                        + str(chan)
                        + ")"
                    )

        # Setup scan list
        msg = "ROUT:SCAN (@"
        for i in range(numChans):
            msg = msg + str(channelList[i])
            if i < (numChans - 1):
                msg = msg + ","
            else:
                msg = msg + ")"
        await self.interface.awrite(msg)

        # Setup trigger and timer & Format
        if samplesPerChan > 1:
            await self.interface.awrite("TRIG:SOUR TIM")
            await self.interface.awrite("TRIG:TIM " + str(timeInterval))
            await self.interface.awrite("TRIG:COUN " + str(samplesPerChan))
            await self.interface.awrite("FORM:READ:TIME ON")
        else:
            await self.interface.awrite("TRIG:SOUR IMM")
            await self.interface.awrite("TRIG:COUN 1")
            await self.interface.awrite("FORM:READ:TIME OFF")
        await self.interface.awrite("FORM:READ:ALAR OFF")
        await self.interface.awrite("FORM:READ:CHAN OFF")
        await self.interface.awrite("FORM:READ:UNIT OFF")

        # Prepare status and event register
        await self.aclear_status()  # *CLS
        await self.event_status_enable_register.aset(1)  # *ESE 1
        await self.status_enable_register.aset(32)  # *SRE 32

        # Initiate scan and trigger Operation Complete event after completion
        await self.interface.awrite("INIT")
        if verbose:
            print(
                datetime.now().isoformat().replace("T", " ")
                + " - Acquisition initiated"
            )

        # Wait for Service Request (triggered by *OPC after the scan
        # is complete)
        await self.await_till_completion_of_operations()  # *OPC
        if sampleRate:
            tmo = 2 * int(1000 * samplesPerChan / sampleRate)
        else:
            tmo = 10000
        print("tmo =", tmo, "ms")
        await self.interface.await_for_srq(timeout=tmo)

        # Unassert SRQ
        await self.aclear_status()

        # Fetch data
        if verbose:
            print(datetime.now().isoformat().replace("T", " ") + " - Fetching data")
        data = await self.interface.aquery("FETCH?", verbose=verbose)

        # Parse data
        if samplesPerChan > 1:
            # timeStamp + value for each channel
            values = np.array([float(x) for x in data.split(",")])
            retval = values[::2], values[1::2]
        else:
            # expectedEntriesPerLine = numChans
            # only value for each channel
            retval = np.array([float(x) for x in data.split(",")])

        return retval


afeatures = [
    AsyncAgilent34970aValue("vdc", doc="DC Voltage", function_name="VOLT:DC"),
    AsyncAgilent34970aValue("vrms", doc="RMS Voltage", function_name="VOLT:AC"),
    AsyncAgilent34970aValue("temperature", doc="Temperature", function_name="TEMP"),
    AsyncAgilent34970aValue("ohm", doc="2-wire resistance", function_name="RES"),
    AsyncAgilent34970aValue("ohm_4w", doc="4-wire resistance", function_name="FRES"),
    AsyncAgilent34970aValue("idc", doc="DC Current", function_name="CURR:DC"),
]


AsyncAgilent34970a._build_class_with_features(afeatures)



"""Asynchronous Cryocon 24c
===========================

.. autoclass:: AsyncCryocon24c
   :members:
   :private-members:
   :show-inheritance:

"""

import asyncio
import fluidlab.instruments.multiplexer.cryocon_24c as fl_cryocon
from pymanip.aioinstruments.aioiec60488 import AsyncIEC60488
from pymanip.aioinstruments.aiosocket import AsyncSocketInstrument


class AsyncCryocon24c(AsyncIEC60488, AsyncSocketInstrument, fl_cryocon.Cryocon24c):
    pass



"""

Function definition for intiating and closing sessions and search sessions:
- NISysCfgInitializeSession
- NISysCfgFindHardware
- NISysCfgNextResource
- NISysCfgCloseHandle

"""

import ctypes

from pymanip.nisyscfg._lib import lib
from pymanip.nisyscfg._lib.types import NISysCfgEnumExpertHandle, NISysCfgSessionHandle
from pymanip.nisyscfg._lib.constants import (
    NISysCfgLocale,
    NISysCfgFilterMode,
    NISysCfgStatus,
)


def NISysCfgInitializeSession(
    target="localhost",
    user=None,
    password=None,
    lang=NISysCfgLocale.Default,
    force_refresh=False,
    connect_timeout=1000,
):
    if isinstance(target, str):
        target = target.encode("ascii")
    if isinstance(user, str):
        user = user.encode("ascii")
    if isinstance(password, str):
        password = password.encode("ascii")
    f = lib.NISysCfgInitializeSession
    f.argtypes = (
        ctypes.c_char_p,
        ctypes.c_char_p,
        ctypes.c_char_p,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_uint,
        ctypes.POINTER(NISysCfgEnumExpertHandle),
        ctypes.POINTER(NISysCfgSessionHandle),
    )
    f.restype = ctypes.c_ulong
    expertHdl = NISysCfgEnumExpertHandle(0)
    sesnHdl = NISysCfgSessionHandle(0)
    status = f(
        target,
        user,
        password,
        lang,
        force_refresh,
        connect_timeout,
        ctypes.byref(expertHdl),
        ctypes.byref(sesnHdl),
    )
    return NISysCfgStatus(status), expertHdl, sesnHdl


def NISysCfgFindHardware(
    sesnHandle,
    filterMode=NISysCfgFilterMode.MatchValuesAll,
    filterHandle=None,
    expert_names=None,
):
    f = lib.NISysCfgFindHardware
    f.argtypes = (
        NISysCfgSessionHandle,
        ctypes.c_int,
        ctypes.c_void_p,
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_void_p),
    )
    f.restype = ctypes.c_ulong
    resourceEnumHandle = ctypes.c_void_p(0)
    status = f(
        sesnHandle,
        filterMode,
        filterHandle,
        expert_names,
        ctypes.byref(resourceEnumHandle),
    )
    return NISysCfgStatus(status), resourceEnumHandle


def NISysCfgNextResource(sesnHandle, resourceEnumHandle):
    f = lib.NISysCfgNextResource
    f.argtypes = (
        NISysCfgSessionHandle,
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_void_p),
    )
    f.restype = ctypes.c_ulong
    resourceHandle = ctypes.c_void_p(0)
    status = f(sesnHandle, resourceEnumHandle, ctypes.byref(resourceHandle))
    return NISysCfgStatus(status), resourceHandle


def NISysCfgCloseHandle(handle):
    f = lib.NISysCfgCloseHandle
    f.argtypes = (ctypes.c_void_p,)
    f.restype = ctypes.c_ulong
    status = f(handle)
    return NISysCfgStatus(status)



"""DAQmx acquisition module (:mod:`pymanip.daq.DAQmx`)
======================================================

The :mod:`fluidlab.daq.daqmx` module is a simple functional front-end to
the third-party :mod:`PyDAQmx` module. It mainly provides two simple one-liner functions:

- :func:`~fluidlab.daq.daqmx.read_analog`

- :func:`~fluidlab.daq.daqmx.write_analog`

The :mod:`pymanip.daq.DAQmx` module is essentially based on its fluidlab counterpart, with
other choices for the default values of arguments, and an additionnal `autoset` feature for
the :func:`read_analog` function.
It also adds a convenience function for printing the list of DAQmx devices, used by the
pymanip CLI interface.

A discovery function is also added, :func:`~pymanip.daq.daqmx.print_connected_devices`, based
on the dedicated :class:`~pymanip.daq.daqmx.DAQDevice` class, which is used by the
`list_daq` sub-command on pymanip command line.

.. autoclass:: DAQDevice
   :members:
   :private-members:

.. autofunction:: print_connected_devices

.. autofunction:: read_analog

"""

import ctypes
import numpy as np
from fluiddyn.util.terminal_colors import cprint
import fluidlab.daq.daqmx as daqmx
from fluidlab.daq.daqmx import write_analog


class DAQDevice:
    """This class is represents a DAQmx device.

    :param device_name: name of the DAQmx device, e.g. "Dev1"
    :type device_name: str

    It mostly implement a number of property getters, which are wrappers to the
    :mod:`PyDAQmx` low-level functions.

    In addition, it has a static method, :meth:`~pymanip.daq.DAQmx.DAQDevice.list_connected_devices`
    to discover currently connected devices.
    """

    @staticmethod
    def list_connected_devices():
        """This static method discovers the connected devices.

        :return: connected devices
        :rtype: list of :class:`pymanip.daq.DAQmx.DAQDevice` objects
        """
        try:
            from PyDAQmx import DAQmxGetSystemInfoAttribute, DAQmx_Sys_DevNames

            bufsize = 1024
            buf = ctypes.create_string_buffer(bufsize)
            DAQmxGetSystemInfoAttribute(DAQmx_Sys_DevNames, ctypes.byref(buf), bufsize)
            return [DAQDevice(s.strip().decode("ascii")) for s in buf.value.split(b",")]
        except ImportError:
            print("Cannot list connected devices.")
            return None
            pass

    def __init__(self, device_name):
        """Constructor method
        """
        self.device_name = device_name

    @property
    def product_category(self):
        """Device product category (str)"""
        try:
            from PyDAQmx import (
                DAQmxGetDevProductCategory,
                DAQmx_Val_MSeriesDAQ,
                DAQmx_Val_XSeriesDAQ,
                DAQmx_Val_ESeriesDAQ,
                DAQmx_Val_SSeriesDAQ,
                DAQmx_Val_BSeriesDAQ,
                DAQmx_Val_SCSeriesDAQ,
                DAQmx_Val_USBDAQ,
                DAQmx_Val_AOSeries,
                DAQmx_Val_DigitalIO,
                DAQmx_Val_TIOSeries,
                DAQmx_Val_DynamicSignalAcquisition,
                DAQmx_Val_Switches,
                DAQmx_Val_CompactDAQChassis,
                DAQmx_Val_CSeriesModule,
                DAQmx_Val_SCXIModule,
                DAQmx_Val_SCCConnectorBlock,
                DAQmx_Val_SCCModule,
                DAQmx_Val_NIELVIS,
                DAQmx_Val_NetworkDAQ,
                DAQmx_Val_SCExpress,
                DAQmx_Val_Unknown,
            )

            category = ctypes.c_int32(DAQmx_Val_Unknown)
            DAQmxGetDevProductCategory(self.device_name, ctypes.byref(category))
            return {
                DAQmx_Val_MSeriesDAQ: "M Series DAQ",
                DAQmx_Val_XSeriesDAQ: "X Series DAQ",
                DAQmx_Val_ESeriesDAQ: "E Series DAQ",
                DAQmx_Val_SSeriesDAQ: "S Series DAQ",
                DAQmx_Val_BSeriesDAQ: "B Series DAQ",
                DAQmx_Val_SCSeriesDAQ: "SC Series DAQ",
                DAQmx_Val_USBDAQ: "USB DAQ",
                DAQmx_Val_AOSeries: "AO Series",
                DAQmx_Val_DigitalIO: "Digital I/O",
                DAQmx_Val_TIOSeries: "TIO Series",
                DAQmx_Val_DynamicSignalAcquisition: "Dynamic Signal Acquisition",
                DAQmx_Val_Switches: "Switches",
                DAQmx_Val_CompactDAQChassis: "CompactDAQ chassis",
                DAQmx_Val_CSeriesModule: "C Series I/O module",
                DAQmx_Val_SCXIModule: "SCXI module",
                DAQmx_Val_SCCConnectorBlock: "SCC Connector Block",
                DAQmx_Val_SCCModule: "SCC Module",
                DAQmx_Val_NIELVIS: "NI ELVIS",
                DAQmx_Val_NetworkDAQ: "Network DAQ",
                DAQmx_Val_SCExpress: "SC Express",
                DAQmx_Val_Unknown: "Unknown by DAQmx",
            }.get(category.value, "Unknown")
        except ImportError:
            return None

    @property
    def product_type(self):
        """Device product type"""
        from PyDAQmx import DAQmxGetDevProductType

        bufsize = 1024
        buf = ctypes.create_string_buffer(bufsize)
        DAQmxGetDevProductType(self.device_name, buf, bufsize)
        return buf.value.decode("ascii")

    @property
    def product_num(self):
        """Device product num"""
        from PyDAQmx import DAQmxGetDevProductNum

        num = ctypes.c_uint32(0)
        DAQmxGetDevProductNum(self.device_name, ctypes.byref(num))
        return num.value

    @property
    def ai_chans(self):
        """List of the analog input channels on the device"""
        from PyDAQmx import DAQmxGetDevAIPhysicalChans

        bufsize = 2048
        buf = ctypes.create_string_buffer(bufsize)
        DAQmxGetDevAIPhysicalChans(self.device_name, buf, bufsize)
        chans = [s.strip().decode("ascii") for s in buf.value.split(b",")]
        if chans == [""]:
            chans = []
        return chans

    @property
    def ao_chans(self):
        """List of the analog output channels on the device"""
        from PyDAQmx import DAQmxGetDevAOPhysicalChans

        bufsize = 2048
        buf = ctypes.create_string_buffer(bufsize)
        DAQmxGetDevAOPhysicalChans(self.device_name, buf, bufsize)
        chans = [s.strip().decode("ascii") for s in buf.value.split(b",")]
        if chans == [""]:
            chans = []
        return chans

    @property
    def di_lines(self):
        """List of digital input lines on the device"""
        from PyDAQmx import DAQmxGetDevDILines

        bufsize = 2048
        buf = ctypes.create_string_buffer(bufsize)
        DAQmxGetDevDILines(self.device_name, buf, bufsize)
        chans = [s.strip().decode("ascii") for s in buf.value.split(b",")]
        if chans == [""]:
            chans = []
        return chans

    @property
    def di_ports(self):
        """List of digital input ports on the device"""
        from PyDAQmx import DAQmxGetDevDIPorts

        bufsize = 2048
        buf = ctypes.create_string_buffer(bufsize)
        DAQmxGetDevDIPorts(self.device_name, buf, bufsize)
        chans = [s.strip().decode("ascii") for s in buf.value.split(b",")]
        if chans == [""]:
            chans = []
        return chans

    @property
    def do_lines(self):
        """List of digital output lines on the device"""
        from PyDAQmx import DAQmxGetDevDOLines

        bufsize = 2048
        buf = ctypes.create_string_buffer(bufsize)
        DAQmxGetDevDOLines(self.device_name, buf, bufsize)
        chans = [s.strip().decode("ascii") for s in buf.value.split(b",")]
        if chans == [""]:
            chans = []
        return chans

    @property
    def do_ports(self):
        """List of digital output ports on the device"""
        from PyDAQmx import DAQmxGetDevDOPorts

        bufsize = 2048
        buf = ctypes.create_string_buffer(bufsize)
        DAQmxGetDevDOPorts(self.device_name, buf, bufsize)
        chans = [s.strip().decode("ascii") for s in buf.value.split(b",")]
        if chans == [""]:
            chans = []
        return chans

    @property
    def bus_type(self):
        """Bus type connection to the device"""
        from PyDAQmx import (
            DAQmxGetDevBusType,
            DAQmx_Val_PCI,
            DAQmx_Val_PXI,
            DAQmx_Val_SCXI,
            DAQmx_Val_PCCard,
            DAQmx_Val_USB,
            DAQmx_Val_Unknown,
        )

        t = ctypes.c_int32(0)
        DAQmxGetDevBusType(self.device_name, ctypes.byref(t))
        return {
            DAQmx_Val_PCI: "PCI",
            DAQmx_Val_PXI: "PXI",
            DAQmx_Val_SCXI: "SCXI",
            DAQmx_Val_PCCard: "PCCard",
            DAQmx_Val_USB: "USB",
            DAQmx_Val_Unknown: "DAQmx unknown",
        }.get(t.value, "Unknown")

    @property
    def pci_busnum(self):
        """PCI Bus number"""
        from PyDAQmx import DAQmxGetDevPCIBusNum

        num = ctypes.c_uint32(0)
        DAQmxGetDevPCIBusNum(self.device_name, ctypes.byref(num))
        return num.value

    @property
    def pci_devnum(self):
        """PCI Device number"""
        from PyDAQmx import DAQmxGetDevPCIDevNum

        num = ctypes.c_uint32(0)
        DAQmxGetDevPCIDevNum(self.device_name, ctypes.byref(num))
        return num.value

    @property
    def pxi_chassisnum(self):
        """PXI Chassis number"""
        from PyDAQmx import DAQmxGetDevPXIChassisNum

        num = ctypes.c_uint32(0)
        DAQmxGetDevPXIChassisNum(self.device_name, ctypes.byref(num))
        return num.value

    @property
    def pxi_slotnum(self):
        """PXI Slot number"""
        from PyDAQmx import DAQmxGetDevPXISlotNum

        num = ctypes.c_uint32(0)
        DAQmxGetDevPXISlotNum(self.device_name, ctypes.byref(num))
        return num.value

    @property
    def location(self):
        """Description of the location (PCI bus and number, or PXI chassis and slot)"""
        from PyDAQmx import DAQError

        bus = self.bus_type
        if bus == "PCI":
            desc = "PCI {:d}, {:d}".format(self.pci_busnum, self.pci_devnum)
        elif bus == "PXI":
            try:
                desc = "PXI chassis {:d} slot {:d}".format(
                    self.pxi_chassisnum, self.pxi_slotnum
                )
            except DAQError:
                # Si le chassis n'est pas identifié alors DAQmx ne peut pas
                # renvoyer les informations, et une exception est levée
                desc = "PXI (unidentified)"
                pass
        else:
            desc = bus
        return desc


def print_connected_devices():
    """This function prints the list of connected DAQmx devices.
    """
    for device in DAQDevice.list_connected_devices():
        print(
            "**",
            device.device_name,
            "(" + device.product_type + ") on",
            device.location,
            "**",
        )
        print("Analog input  :", device.ai_chans)
        print("Analog output :", device.ao_chans)


# Ici read_analog est verbose=True par défaut contrairement à fluidlab
# et on ajoute une fonction "autoset" si volt_min, volt_max sont None
def read_analog(
    resource_names,
    terminal_config,
    volt_min=None,
    volt_max=None,
    samples_per_chan=1,
    sample_rate=1,
    coupling_types="DC",
    output_filename=None,
    verbose=True,
):
    """This function reads signal from analog input.

    :param resources_names: names from MAX (Dev1/ai0)
    :type resource_names: str, or list
    :param terminal_config: "Diff", "RSE", "NRSE"
    :type terminal_config: str, or list
    :param volt_min: minimum voltage
    :type volt_min: float, or list, optional
    :param volt_max: maximum voltage
    :type volt_max: float, or list, optional
    :param samples_per_chan: Number of samples to be read per channel
    :type samples_per_chan: int
    :param sample_rate: Clock frequency
    :type sample_rate: float
    :param coupling_type: Coupling of the channels ("DC", "AC", "GND")
    :type coupling_type: str, or list
    :param output_filename: If not None, file to write the acquired data
    :type output_filename: str, optional
    :param verbose: Verbosity level. Defaults to True (unlike in Fluidlab)
    :type verbose: bool, optional

    If the channel range is not specified, a 5.0 seconds samples will first be acquired
    to determine appropriate channel range (autoset feature).

    """

    # Les type checks ci-dessous ne sont pas très pythoniques
    # mais nécessaire parce que PyDAQmx est une passerelle vers le C
    # et il y a des sous-entendus de type.

    # Ensure that samples_per_chan is integer
    if not isinstance(samples_per_chan, int):
        samples_per_chan = int(samples_per_chan)

    # Ensure resource_names is str or list of str
    if isinstance(resource_names, str):
        num_channels = 1
        resource_names = str(resource_names)
    else:
        num_channels = len(resource_names)
        resource_names = [str(r) for r in resource_names]

    # If no range is provided, take a 5s sample
    if volt_min is None or volt_max is None:
        print("Sampling 5s data to determine channel range")
        if num_channels == 1:
            volt_min = -10.0
            volt_max = 10.0
        else:
            volt_min = [-10.0] * num_channels
            volt_max = [10.0] * num_channels

        data = daqmx.read_analog(
            resource_names,
            terminal_config,
            volt_min,
            volt_max,
            samples_per_chan=50000,
            sample_rate=10e3,
            coupling_types=coupling_types,
            verbose=False,
        )
        if num_channels == 1:
            volt_range = np.max(np.abs(data)) * 1.25
            volt_min = -volt_range
            volt_max = volt_range
        else:
            for chan in range(num_channels):
                volt_range = np.max(np.abs(data[chan])) * 1.25
                volt_min[chan] = -volt_range
                volt_max[chan] = volt_range
                print(
                    "Channel", chan, "min max:", np.min(data[chan]), np.max(data[chan])
                )

    # Run fluidlab daqmx.read_analog with verbose=True by default
    data = daqmx.read_analog(
        resource_names,
        terminal_config,
        volt_min,
        volt_max,
        samples_per_chan,
        sample_rate,
        coupling_types,
        output_filename,
        verbose,
    )

    # If verbose, check that voltage range has not been reached and issue a warning otherwise
    if verbose:
        if num_channels == 1:
            channel_range = np.max([np.abs(volt_min), np.abs(volt_max)])
            if np.max(np.abs(data)) >= channel_range:
                cprint.red("WARNING: channel range too small!")
        else:
            for chan in range(num_channels):
                try:
                    channel_range = np.max(
                        [np.abs(volt_min[chan]), np.abs(volt_max[chan])]
                    )
                except TypeError:
                    channel_range = np.max([np.abs(volt_min), np.abs(volt_max)])
                    pass

                if np.max(np.abs(data[chan])) >= channel_range:
                    cprint.red(
                        "WARNING: channel range is too small for channel "
                        + resource_names[chan]
                    )

    return data


# def write_analog(resource_names, sample_rate, signals, blocking=True):
#    return daqmx.write_analog(resource_names, sample_rate, signals, blocking)


def write_analog_end_task(task, timeout=0.0):
    daqmx.write_analog_end_task(task, timeout)


def measure_freq(resource_name, freq_min=1, freq_max=1000):
    return daqmx.measure_freq(resource_name, freq_min, freq_max)



import asyncio


def synchronize_generator(async_generator, *args, **kwargs):
    """
    Returns a synchronous generator from an asynchronous generator
    """

    ag = async_generator(*args, **kwargs)

    async def consume_generator(stop_signal):
        r = await ag.asend(stop_signal)
        return r

    loop = asyncio.get_event_loop()
    try:
        stop_signal = None
        while not stop_signal:
            val = loop.run_until_complete(consume_generator(stop_signal))
            stop_signal = yield val
        if stop_signal:
            val = loop.run_until_complete(consume_generator(stop_signal))
            yield val
    except StopAsyncIteration:
        pass


def synchronize_function(async_func, *args, **kwargs):
    """
    Execute synchronously an asynchronous function
    """

    try:
        r = asyncio.run(async_func(*args, **kwargs))
    except AttributeError:
        loop = asyncio.get_event_loop()
        r = loop.run_until_complete(async_func(*args, **kwargs))
    return r


if __name__ == "__main__":

    async def spam(n):
        for i in range(n):
            yield i ** n
            await asyncio.sleep(1.0)

    for x in synchronize_generator(spam, 3):
        print(x)

    async def f(x):
        for i in range(x):
            print(i)
            await asyncio.sleep(0.5)
        return i

    a = synchronize_function(f, 5)
    b = synchronize_function(f, 3)
    print(b)

    def sync_spam(n):
        yield from synchronize_generator(spam, n)

    for s in sync_spam(5):
        print(s)



"""

NI System Configuration constants

"""

from enum import IntEnum

__all__ = [
    "NISYSCFG_SIMPLE_STRING_LENGTH",
    "NISysCfgLocale",
    "NISysCfgFilterMode",
    "NISysCfgResourceProperty",
    "NISysCfgIndexedProperty",
    "NISysCfgStatus",
]

NISYSCFG_SIMPLE_STRING_LENGTH = 1024


class NISysCfgLocale(IntEnum):
    Default = (0,)
    ChineseSimplified = 2052
    English = 1033
    French = 1036
    German = 1031
    Japanese = 1041
    Korean = 1042


class NISysCfgFilterMode(IntEnum):
    MatchValuesAll = 1
    MatchValuesAny = 2
    MatchValuesNone = 3
    ModeAllPropertiesExist = 4


class NISysCfgResourceProperty(IntEnum):
    # Read-only properties
    IsDevice = 16781312  # NISysCfgBool
    IsChassis = 16941056  # NISysCfgBool
    ConnectsToBusType = 16785408  # NISysCfgBusType
    VendorId = 16789504  # ctypes.c_uint,
    VendorName = 16793600  # ctypes.c_char_p,
    ProductId = 16797696  # ctypes.c_uint,
    ProductName = 16801792  # ctypes.c_char_p,
    SerialNumber = 16805888  # ctypes.c_char_p,
    FirmwareRevision = 16969728  # ctypes.c_char_p,
    IsNIProduct = 16809984  # NISysCfgBool
    IsSimulated = 16814080  # NISysCfgBool
    ConnectsToLinkName = 16818176  # ctypes.c_char_p,
    HasDriver = 16920576  # NISysCfgHasDriverType
    IsPresent = 16924672  # NISysCfgIsPresentType
    SlotNumber = 16822272  # int
    SupportsInternalCalibration = 16842752  # NISysCfgBool
    InternalCalibrationLastTime = 16846848  # NISysCfgTimestampUTC
    InternalCalibrationLastTemp = 16850944  # double
    SupportsExternalCalibration = 16859136  # NISysCfgBool
    ExternalCalibrationLastTemp = 16867328  # double
    CalibrationComments = 16961536  # ctypes.c_char_p,
    CurrentTemp = 16965632  # double
    PxiPciBusNumber = 16875520  # ctypes.c_uint,
    PxiPciDeviceNumber = 16879616  # ctypes.c_uint,
    PxiPciFunctionNumber = 16883712  # ctypes.c_uint,
    PxiPciLinkWidth = 16973824  # int
    PxiPciMaxLinkWidth = 16977920  # int
    UsbInterface = 16887808  # ctypes.c_uint,
    TcpHostName = 16928768  # ctypes.c_char_p,
    TcpMacAddress = 16986112  # ctypes.c_char_p,
    TcpIpAddress = 16957440  # ctypes.c_char_p,
    TcpDeviceClass = 17022976  # ctypes.c_char_p,
    GpibPrimaryAddress = 16994304  # int
    GpibSecondaryAddress = 16998400  # int
    ProvidesBusType = 16932864  # NISysCfgBusType
    ProvidesLinkName = 16936960  # ctypes.c_char_p,
    NumberOfSlots = 16826368  # int
    SupportsFirmwareUpdate = 17080320  # NISysCfgBool
    FirmwareFilePattern = 17084416  # ctypes.c_char_p,
    RecommendedCalibrationInterval = 17207296  # int
    SupportsCalibrationWrite = 17215488  # NISysCfgBool
    HardwareRevision = 17256448  # ctypes.c_char_p,
    CpuModelName = 17313792  # ctypes.c_char_p,
    CpuSteppingRevision = 17317888  # int

    # Read/Write firmware properties
    FirmwareUpdateMode = 17354752  # NISysCfgFirmwareUpdateMode

    # Read/Write calibration properties
    ExternalCalibrationLastTime = 16863232  # NISysCfgTimestampUTC
    RecommendedNextCalibrationTime = 16871424  # NISysCfgTimestampUTC

    # Write-only calibration properties
    CalibrationCurrentPassword = 17223680  # ctypes.c_char_p,
    CalibrationNewPassword = 17227776  # ctypes.c_char_p,

    # Read/Write remote access properties
    SysCfgAccess = 219504640  # NISysCfgAccessType

    # Read-only network adapter properties
    AdapterType = 219332608  # NISysCfgAdapterType
    MacAddress = 219168768  # ctypes.c_char_p,

    # Read/Write network adapter properties
    AdapterMode = 219160576  # NISysCfgAdapterMode
    TcpIpRequestMode = 219172864  # NISysCfgIpAddressMode
    TcpIpv4Address = 219181056  # ctypes.c_char_p,
    TcpIpv4Subnet = 219189248  # ctypes.c_char_p,
    TcpIpv4Gateway = 219193344  # ctypes.c_char_p,
    TcpIpv4DnsServer = 219197440  # ctypes.c_char_p,
    TcpPreferredLinkSpeed = 219213824  # NISysCfgLinkSpeed
    TcpCurrentLinkSpeed = 219222016  # NISysCfgLinkSpeed
    TcpPacketDetection = 219258880  # NISysCfgPacketDetection
    TcpPollingInterval = 219262976  # ctypes.c_uint,
    IsPrimaryAdapter = 219308032  # NISysCfgBool
    EtherCatMasterId = 219250688  # ctypes.c_uint,
    EtherCatMasterRedundancy = 219500544  # NISysCfgBool

    # Read-only wireless network adapter properties
    WlanBssid = 219398144  # ctypes.c_char_p,
    WlanCurrentLinkQuality = 219394048  # ctypes.c_uint,

    # Read/Write wireless network adapter properties
    WlanCurrentSsid = 219377664  # ctypes.c_char_p,
    WlanCurrentConnectionType = 219381760  # NISysCfgConnectionType
    WlanCurrentSecurityType = 219385856  # NISysCfgSecurityType
    WlanCurrentEapType = 219389952  # NISysCfgEapType
    WlanCountryCode = 219406336  # int
    WlanChannelNumber = 219410432  # ctypes.c_uint,
    WlanClientCertificate = 219422720  # ctypes.c_char_p,

    # Write-only wireless network adapter properties
    WlanSecurityIdentity = 219414528  # ctypes.c_char_p,
    WlanSecurityKey = 219418624  # ctypes.c_char_p,

    # Read-only time properties
    SystemStartTime = 17108992  # NISysCfgTimestampUTC

    # Read/Write time properties
    CurrentTime = 219279360  # NISysCfgTimestampUTC
    TimeZone = 219471872  # ctypes.c_char_p,

    # Read/Write startup settings properties
    UserDirectedSafeModeSwitch = 219537408  # NISysCfgBool
    ConsoleOutSwitch = 219541504  # NISysCfgBool
    IpResetSwitch = 219545600  # NISysCfgBool

    # Read-only counts for indexed properties
    NumberOfDiscoveredAccessPoints = 219365376  # ctypes.c_uint,
    NumberOfExperts = 16891904  # int
    NumberOfServices = 17010688  # int
    NumberOfAvailableFirmwareVersions = 17088512  # int
    NumberOfCpus = 17137664  # int
    NumberOfFans = 17174528  # int
    NumberOfTemperatureSensors = 17186816  # int
    NumberOfVoltageSensors = 17149952  # int
    NumberOfUserLedIndicators = 17281024  # int
    NumberOfUserSwitches = 17293312  # int


class NISysCfgIndexedProperty(IntEnum):
    ExpertName = 16900096
    ExpertResourceName = 16896000
    ExpertUserAlias = 16904192


class NISysCfgStatus(IntEnum):
    # Common success codes
    OK = 0  # The operation succeeded.
    EndOfEnum = (
        1  # Reached end of the enumeration. Used by the NISysCfgNext* functions.
    )
    SelfTestBasicOnly = (
        0x00040370
    )  # The expert performed a basic self-test because it does not implement the specified mode.
    FoundCachedOfflineSystem = (
        0x00040400
    )  # Initialization succeeded but the target is offline. Only cached system properties are available.
    RestartLocalhostInitiated = (
        0x00040401
    )  # For the local system, the option to wait until the restart is complete is ignored. The function has successfully initiated a restart with the operating system.

    # Common error codes
    NotImplemented = (
        0x80004001  # This operation is not implemented for this target or resource.
    )
    NullPointer = 0x80004003  # A required pointer parameter was NULL.
    Fail = 0x80004005  # Miscellaneous operation failure.
    Unexpected = (
        0x8000FFFF
    )  # A critical unexpected error occurred. Please report this to National Instruments.
    OutOfMemory = 0x8007000E  # Out of memory.
    InvalidArg = 0x80070057  # Some parameter is invalid.
    OperationTimedOut = 0x80040420  # The operation timed out.
    FileNotFound = 0x8004049E  # The specified file was not found.
    InvalidMACFormat = (
        0x800404CA
    )  # Unsupported MAC address format. Supply the MAC address as a colon separated string of characters instead of Hex display.

    # 'Hardware' and 'Resource' functions
    PropMismatch = (
        0x80040370  # The property already exists with a different type or value.
    )
    PropDoesNotExist = 0x80040371  # The property does not exist for this resource.
    UriIllegalSyntax = (
        0x80040372
    )  # The name of the target or expert contains illegal characters or has an invalid format. Each label of the hostname must be between 1 and 63 characters long, and the entire hostname, including delimiting dots, must be 255 characters or less.
    UriTargetDoesNotExist = (
        0x80040373
    )  # Could not contact the NI System Configuration API at the specified target address. Ensure that the system is online.
    UriExpertDoesNotExist = 0x80040374  # A specified expert is not installed.
    ItemDoesNotExist = 0x80040375  # The specified resource name does not exist.
    InvalidMode = 0x80040376  # The specified mode is invalid.
    SysConfigAPINotInstalled = (
        0x80040378
    )  # The NI System Configuration API is not installed on the specified target.
    NameSyntaxIllegal = 0x8004037A  # The suggested name contains illegal characters.
    NameCollision = 0x8004037B  # Another resource already has the suggested name.
    NoPropValidated = 0x8004037C  # None of the changed properties could be validated.
    UriUnauthorized = (
        0x8004037D
    )  # The current user does not have permission for the requested operation.
    RenameResourceDependencies = (
        0x8004037E
    )  # The resource being renamed has dependencies, and the 'updateDependencies' flag was false.
    ValueInvalid = (
        0x8004037F  # A property contained a value that is not valid or is out of range.
    )
    ValuesInconsistent = (
        0x80040380
    )  # Multiple properties contained values that are inconsistent with each other.
    Canceled = 0x80040381  # The operation was canceled.
    ResponseSyntax = (
        0x80040382
    )  # Could not parse the response from the NI System Configuration API at the specified target address.
    ResourceIsNotPresent = (
        0x80040383
    )  # The resource name is valid but the operation requires the resource to be present.
    ResourceIsSimulated = (
        0x80040384
    )  # The resource name is valid but the operation is not supported on simulated resources.
    NotInFirmwareUpdateState = (
        0x80040385
    )  # The resource requires being in the firmware update state to perform this operation.
    FirmwareImageDeviceMismatch = (
        0x80040386  # The uploaded firmware image does not work with this resource.
    )
    FirmwareImageCorrupt = (
        0x80040387  # The uploaded firmware image is corrupt or incomplete.
    )
    InvalidFirmwareVersion = (
        0x80040388  # The specified firmware version does not exist.
    )
    OlderFirmwareVersion = (
        0x80040389
    )  # The specified firmware version is older than what is currently installed.
    InvalidLoginCredentials = 0x8004038A  # The username or password is incorrect.
    FirmwareUpdateAttemptFailed = (
        0x8004038B
    )  # The specified firmware was not successfully installed. See the output parameters for more information.
    EncryptionFailed = 0x8004038C  # The data could not be encrypted.
    SomePropsNotValidated = (
        0x8004038D
    )  # The changes were not saved. Some of the modified properties were not validated because they do not apply to this item.
    InvalidCalibrationCredentials = 0x8004038E  # The calibration password is incorrect.
    CannotDeletePresentResource = (
        0x8004038F  # Could not delete the specified resource because it is present.
    )
    UriTargetTransmitError = (
        0x80040390
    )  # Failed transmitting data to or from the web server at the specified target address.
    DecryptionFailed = (
        0x80040391
    )  # The NI System Configuration API at the specified target address could not decrypt the data.
    FirmwareExpertVersionMismatch = (
        0x80040392
    )  # The specified firmware requires a newer version of the expert than what is currently installed.
    AmbiguousImportAction = (
        0x80040393
    )  # There was uncertainty regarding what action to take during an import.
    RequiredItemFailedImport = 0x80040394  # A required item could not be imported.

    # 'Report' functions
    PermissionDenied = (
        0x800403B0  # Unable to write to file or folder. Permission denied.
    )
    SystemNotFound = (
        0x800403B1
    )  # Unable to connect to the specified system. Ensure that the system is online.
    TransformFailed = 0x800403B2  # Error running transform to generate report.
    NotInstalled = 0x800403B3  # Unable to find MAX on the system. Please reinstall.
    LaunchFailure = 0x800403B4  # Unexpected error launching nimax.exe.
    InternalTimeout = (
        0x800403B5  # Launched nimax.exe but it did not complete in a reasonable time.
    )
    MissingTransform = (
        0x800403B6  # Unable to find an XSL transform to generate the report.
    )
    IncorrectExtension = 0x800403B7  # Incorrect report file extension provided.
    FileReadOnly = 0x800403B8  # Report file is read-only. Unable to generate report.
    ReportOverwrite = (
        0x800403B9
    )  # Report file exists and the NIMAX_FailIfOverwritingReport flag was set.
    DirectoryError = 0x800403BA  # Error creating directory for report files.

    # 'Export' and 'Import' functions
    CannotOpenFile = 0x80040400  # Error opening a file.
    InsufficientPermissions = (
        0x80040401  # The object cannot be accessed because of insufficient permissions.
    )
    NCECopierFailed = 0x80040402  # Error with the object copier.
    FileOperationFailed = 0x80040403  # Error performing a file operation.
    NameCollisionError = (
        0x80040404  # Names from one expert have collided with another expert.
    )
    UnexpectedError = 0x80040405  # Unexpected error has occurred.
    NCENoStreamError = (
        0x80040406
    )  # The expert requested its stream for import but one does not exist because it did not export a stream.
    NCECompressionError = 0x80040407  # Error compressing or decompressing file.
    NCEStreamReadError = 0x80040408  # Error reading from a stream.
    NCEStreamWriteError = 0x80040409  # Error writing to a stream.
    NCEStreamSeekError = 0x8004040A  # Error seeking to a position in a stream.
    NCERepoNotReady = 0x8004040B  # Repository not ready to be exported.
    NCERepoInvalid = (
        0x8004040C
    )  # The file or stream from which to import the repository is not a valid repository.
    NCERepoIncompat = (
        0x8004040D
    )  # The repository was exported with a newer version of MAX than what is on the importing machine.
    NCENoImportStorage = 0x8004040E  # The import storage could not be opened.
    NCENoExportStorage = 0x8004040F  # The export storage could not be created.
    NCENoObjCopier = 0x80040410  # The object copier could not be created.
    CopyInProgress = 0x80040411  # A PortCfg operation is already in progress.
    FileNotRecognized = 0x80040412  # The custom file does not belong to a given expert.
    SystemNotSupported = (
        0x80040413  # A specified system is not supported by this expert.
    )
    SystemNotReachable = (
        0x80040414
    )  # A specified system is presumably supported, but network errors prevent connection.
    ProductSoftwareNotInstalled = (
        0x80040415  # The product is not installed on the specified system.
    )
    ProductSoftwareTooOld = (
        0x80040416  # The product is installed on the remote system, but is too old.
    )
    ProductSoftwareTooNew = (
        0x80040417  # The product is installed on the remote system, but is too new.
    )
    DataTooOld = (
        0x80040418
    )  # The import data is too old. The product is not backward-compatible with this data.
    DataTooNew = (
        0x80040419
    )  # The import data is too new. The product is not forward-compatible with this data.
    NoItemsToCopy = (
        0x8004041A  # The operation failed because no source items were specified.
    )
    OrphanItems = 0x8004041B  # The operation failed because some items were orphans.
    DirtyItems = (
        0x8004041C
    )  # The operation failed because some items were in-edit and not saved.
    FileOverwrite = (
        0x8004041D  # The operation failed because it would overwrite a file.
    )
    ItemOverwrite = 0x8004041E  # The operation failed because it would overwrite items.
    MissingDependency = (
        0x8004041F  # The operation failed because of missing dependency items.
    )
    OperationCanceled = 0x80040421  # The operation was canceled by the client.
    WarningConflicts = 0x80040422  # The operation failed because of warning conflicts.
    ErrorConflicts = 0x80040423  # The operation failed because of general conflicts.
    ItemsRequireUserInput = (
        0x80040424
    )  # The operation failed because of unresolved conflicts requiring user input.
    ProductExpertNotReady = (
        0x80040425
    )  # An expert is not ready to accept the specified source or destination, but may become ready in the future.
    OrphanFiles = 0x80040426  # The operation failed because some files were orphans.
    IsConst = (
        0x80040427
    )  # Caller called a non-const method on an object that is logically const.
    UnsupportedProductMode = (
        0x80040428
    )  # An expert does not support the attempted copy mode (e.g. merge to file, etc.).

    # 'System' functions
    RestartLocalhostAmbiguous = (
        0x8004046C
    )  # To reboot your system, either specify 'localhost' on the front panel for Session in, or call Initialize Session first.
    ImageInvalidCorrupt = (
        0x8004046D  # The image is corrupt or the file type is invalid.
    )
    SafeOrInstallModeRequired = (
        0x8004046E
    )  # Can only perform this action in safe or install mode, and the 'auto restart' flag was false.
    EncryptPhraseMismatch = (
        0x8004046F
    )  # The encryption passphrase when applying an image was not the same as when the image was created.
    InvalidIP = 0x80040470  # IP Address is invalid.
    InvalidGateway = 0x80040471  # Gateway is invalid.
    InvalidDNS = 0x80040472  # DNS is invalid.
    InvalidSubnet = 0x80040473  # Subnet is invalid.
    CmdNotSupported = 0x80040474  # Command is not supported by given protocol.
    ConfigFailed = (
        0x80040475  # Remote system replied with the failure to config command.
    )
    Locked = 0x80040476  # Remote system is locked. Requires a password to configure.
    BadPassword = 0x80040477  # The password supplied for the operation is invalid.
    NotConfigurable = (
        0x80040478
    )  # The remote device is not configurable for some reason other than password.
    UnlockFailed = 0x80040479  # Failed to unlock the system.
    LockFailed = 0x8004047A  # Failed to lock the system.
    InstallFailed = 0x8004047B  # General installation failure.
    InstallationCorrupt = (
        0x8004047C
    )  # Installation component files were not found in the repository, or were corrupt.
    EmptyFile = 0x8004047D  # The installation file is empty.
    UnconfiguredIP = (
        0x8004047E
    )  # The system must have a valid IP before certain operations such as installation. Cannot be 0.0.0.0.
    InstallationGenericFailure = 0x80040480  # General install error.
    DownloadAlreadyStarted = (
        0x80040482
    )  # Installation to the specified target has already started. Multiple simultaneous installations are not allowed.
    Aborted = 0x80040483  # Remote action aborted.
    DiskFull = (
        0x8004048E
    )  # Hard Drive on the remote system is either full or has encountered an I/O error.
    HDFormatFailed = 0x8004048F  # Hard Drive Format failed.
    HDFormatNotSafeMode = (
        0x80040490  # System must be in safe mode before attempting Hard Drive Format.
    )
    HDFormatRebootFailed = (
        0x80040491
    )  # System failed to reboot after Hard Drive format. System is in unknown state.
    ConnectionRefused = 0x80040492  # The server refused the network connection.
    GetRemoteFilesFailed = (
        0x80040495  # Failed to get one or more files while creating system image.
    )
    PutRemoteFilesFailed = (
        0x80040496  # Failed to put one or more files while applying system image.
    )
    InvalidImage = 0x80040497  # The specified path does not point to a valid image.
    ImageDeviceCodeMismatch = (
        0x80040498
    )  # The image is for a different device class and is incompatible with the target.
    SystemMismatch = (
        0x80040499  # The image was not originally created from the specified target.
    )
    HDFormatWrongFS = (
        0x8004049A
    )  # The requested file system is not supported on the specified target.
    CustomInstallNotSupported = (
        0x8004049B
    )  # The specified target does not support custom software installations.
    FTPFailed = 0x8004049C  # A file transfer error (FTP or WebDAV) occurred.
    Timeout = 0x8004049D  # Operation timed out.
    DirNotFound = 0x8004049F  # The specified directory was not found.
    PathNotFound = 0x800404A0  # The specified file or directory path was not found.
    NoSoftwareAvailable = 0x800404A1  # No software is available for install.
    OverwriteError = (
        0x800404A2  # The file or directory exists and the overwrite flag was false.
    )
    HDFormatCannotKeepCfg = (
        0x800404A3
    )  # The target was not formatted because the option to keep configuration after a format is supported only for targets on the local subnet. To format the target, repeat the operation without requesting to keep the configuration.
    FileOrPathTooLong = (
        0x800404A4  # Filename or pathname is longer than what the server supports.
    )
    DDPInternalTimeout = (
        0x800404A5
    )  # Failed when communicating with the system. This issue is usually caused by a high latency in the network. Refer to KnowledgeBase article 42GH3O00 on ni.com for possible solutions.
    IOPermissionDenied = (
        0x800404A6  # The operation failed because of insufficient permissions.
    )
    PathAlreadyExists = (
        0x800404A7  # The operation failed because the path already exists.
    )
    ExecutionFailure = (
        0x800404A8
    )  # The execution of an external command, script, or application failed.
    DownloadError = (
        0x800404A9  # Failed to download the file from the 'RT Images' repository.
    )
    NetSendFailed = 0x800404AB  # Failed to send command.
    ContactHostDisconnected = (
        0x800404AC  # Could not contact remote target. Ensure that the system is online.
    )
    NetSvcDown = 0x800404AD  # Could not access network.
    NotConfirmed = (
        0x800404AE
    )  # Command was not confirmed. The result of the operation is uncertain.
    HostNotResolved = 0x800404AF  # Hostname could not be resolved by DNS.
    RebootTimeout = 0x800404B0  # Timeout while waiting for reboot. System is offline.
    NoConfirmationFP1600 = (
        0x800404B1
    )  # Sending new configuration operation returned a failure, but might not necessarily have failed.
    DuplicateStartup = 0x800404B4  # Cannot install more than one startup component.
    RemoteInvalidArgument = 0x800404B5  # Invalid argument passed.
    NotUninstallable = (
        0x800404B6  # Cannot uninstall a component. There are dependencies.
    )
    DuplicatesNotAllowed = (
        0x800404B7  # Cannot install multiple packages of the same component.
    )
    NotInstallable = 0x800404B8  # Cannot install a component. There are dependencies.
    WrongDevice = 0x800404B9  # Component will not work on this target.
    WrongOS = 0x800404BA  # Component will not work on this operating system.
    OSVersionTooOld = 0x800404BB  # A BIOS update is required before installing.
    IOError = 0x800404BC  # Cannot open file or folder.
    CorruptConfig = (
        0x800404BD  # Duplicate or missing components on target installation.
    )
    BufferOverflow = 0x800404BE  # Buffer overflow. Size is too small.
    UnsupportedCDFVersion = (
        0x800404BF
    )  # Unsupported version of CDF format. Needs a newer version of MAX or NI System Configuration.
    InvalidStack = 0x800404C0  # Invalid software set.
    IncompleteStack = (
        0x800404C1
    )  # Incompletely specified list of packages. Some hidden dependencies were added.
    StackItemMissing = (
        0x800404C2
    )  # One or more Software Set items could not be found in the repository.
    TopLevelHiddenComponentError = (
        0x800404C3  # There is a top-level hidden component installed.
    )
    InvalidAddon = (
        0x800404C4
    )  # A component was passed in that is not an installable add-on. It may be an unknown ID, a defined item that is not an add-on, a missing add-on, or a non-installable add-on.
    NoRTImagesFolder = (
        0x800404C5  # Could not find or access 'RT Images' repository location.
    )
    NoRTImagesRegistry = 0x800404C6  # Could not read the 'RT Images' registry key.
    NoRTS2CDF = 0x800404C7  # Could not find the rts2cdf conversion utility.
    UnsupportedOS = 0x800404C8  # The operating system is not supported.
    ExactVersionRequired = (
        0x800404C9
    )  # Unspecified version while trying to install exact version of a component.
    InvalidStartup = 0x800404CB  # A component was passed in that is not a startup.



"""Asynchronous Instrument drivers (:mod:`pymanip.aioinstruments.aiodrivers`)
=============================================================================

This module defines a subclass of :class:`fluidlab.instruments.drivers.Driver` where
the QueryInterface attribute is replaced by the corresponding AsyncQueryInterface
instance. The asynchronous context manager is bound to the interface asynchronous context
manager.

.. autofunction:: interface_from_string

.. autoclass:: AsyncDriver
   :members:
   :private-members:


"""

import fluidlab.interfaces as flinter
import fluidlab.instruments.features as flfeatures
import fluidlab.instruments.drivers as fldrivers

from pymanip.interfaces.aiointer import AsyncQueryInterface


def interface_from_string(name, default_physical_interface=None, **kwargs):
    """This function is similar to :func:`fluidlab.interfaces.interface_from_string` except
    that it returns an instance of :class:`AsyncQueryInterface` instead of :class:`QueryInterface`.
    """

    classname, physical_interface = flinter.interface_classname_from_string(
        name, default_physical_interface, **kwargs
    )

    if classname == "VISAInterface":
        from pymanip.interfaces.aiovisa import AsyncVISAInterface

        return AsyncVISAInterface(name, **kwargs)
    elif classname == "SerialInterface":
        from pymanip.interfaces.aioserial import AsyncSerialInterface

        return AsyncSerialInterface(name, **kwargs)
    elif classname == "GPIBInterface":
        from pymanip.interfaces.aiogpib import AsyncGPIBInterface

        if name.startswith("GPIB"):
            board_name, instrument_adress, *others = name.split("::")

            board_adress = int(board_name[4:])
            instrument_adress = int(instrument_adress)

        return AsyncGPIBInterface(board_adress, instrument_adress)

    raise NotImplementedError()


class AsyncDriver(fldrivers.Driver):
    """This class is an asynchronous extension of :class:`fluidlab.instruments.drivers.Driver`.
    """

    def __init__(self, interface=None):
        """Constructor method
        """
        if isinstance(interface, str):
            interface = interface_from_string(
                interface, self.default_physical_interface, **self.default_inter_params
            )
        elif not isinstance(interface, AsyncQueryInterface):
            raise ValueError("Interface should be an AsyncQueryInterface")

        super().__init__(interface)

    async def __aenter__(self):
        """Asynchronous context manager enter method
        """
        await self.interface.__aenter__()
        return self

    async def __aexit__(self, type_, value, cb):
        """Asynchronous context manager exit method
        """
        await self.interface.__aexit__(type_, value, cb)



import platform

__all__ = list()

if platform.system() == "Linux":
    try:
        from fluidlab.interfaces.gpib_inter import GPIBInterface

        __all__.append("GPIBInterface")
    except ImportError:
        print("Warning: Linux-GPIB Python bindings not found")
        pass
    try:
        from fluidlab.interfaces.usbtmc_inter import USBTMCInterface

        __all__.append("USBTMCInterface")
    except ImportError:
        print("Warning: Linux USBTMC bindings not found.")
        pass

try:
    from fluidlab.interfaces.visa_inter import VISAInterface

    __all__.append("VISAInterface")
except ImportError:
    print("Warning: PyVISA not found")
    pass

try:
    from fluidlab.interfaces.serial_inter import SerialInterface

    __all__.append("SerialInterface")
except ImportError:
    print("Warning: SerialInterface not found")
    pass



"""Asynchronous Socket instrument driver (:mod:`pymanip.aioinstruments.aiosocket`)
==================================================================================

.. autoclass:: AsyncSocketInstrument
   :members:
   :private-members:
   :show-inheritance:

"""



"""Asynchronous Instrument features (:mod:`pymanip.aioinstruments.aiofeatures`)
===============================================================================

Asynchronous extension of fluidlab instrument features. The main difference is that they
define :meth:`aget` and :meth:`aset` co-routine methods. The original :meth:`get` and
:meth:`set` are not overridden, and may still be used.

.. autoclass:: AsyncWriteCommand
   :members:
   :private-members:

.. autoclass:: AsyncQueryCommand
   :members:
   :private-members:

.. autoclass:: AsyncValue
   :members:
   :private-members:

.. autoclass:: AsyncNumberValue
   :members:
   :private-members:

.. autoclass:: AsyncFloatValue
   :members:
   :private-members:

.. autoclass:: AsyncBoolValue
   :members:
   :private-members:

.. autoclass:: AsyncIntValue
   :members:
   :private-members:

.. autoclass:: AsyncRegisterValue
   :members:
   :private-members:

"""

import asyncio
import fluidlab.instruments.features as flfeatures


class AsyncWriteCommand(flfeatures.WriteCommand):
    def _build_driver_class(self, Driver):
        super()._build_driver_class(Driver)
        command_str = self.command_str

        async def func(self):
            await self._interface.awrite(command_str)

        func.__doc__ = self.__doc__
        setattr(Driver, self._name, func)


class AsyncQueryCommand(flfeatures.QueryCommand):
    def _build_driver_class(self, Driver):
        super()._build_driver_class(Driver)

        command_str = self.command_str

        if self.parse_result is None:

            async def func(self):
                return await self._interface.aquery(command_str)

        else:

            async def func(self):
                r = await self._interface.aquery(command_str)
                return self.parse_result(r)

        func.__doc__ = self.__doc__
        setattr(Driver, self._name, func)


class AsyncValue(flfeatures.Value):
    async def aget(self, channel=0):
        if isinstance(channel, list) or isinstance(channel, tuple):
            return [await self.aget(c) for c in channel]

        if self.pause_instrument > 0:
            await asyncio.sleep(self.pause_instrument)

        command = self.command_get
        if self.channel_argument:
            command = command.format(channel=channel)
        if self.pause_instrument > 0:
            r = await self._interface.aquery(command, time_delay=self.pause_instrument)
        else:
            r = await self._interface.aquery(command)
        result = self._convert_from_str(r)
        self._check_value(result)
        return result

    async def aset(self, value, channel=0):
        if self.pause_instrument > 0:
            await asyncio.sleep(self.pause_instrument)
        self._check_value(value)
        if self.channel_argument:
            # here we don't call _convert_as_str to allow the user to choose
            # the desired format in the command_set string
            command = self.command_set.format(channel=channel, value=value)
        else:
            command = self.command_set + " " + self._convert_as_str(value)
        await self._interface.awrite(command)
        if self.check_instrument_value_after_set:
            self._check_instrument_value(value)


class AsyncNumberValue(AsyncValue, flfeatures.NumberValue):
    pass


class AsyncFloatValue(AsyncNumberValue, flfeatures.FloatValue):
    pass


class AsyncBoolValue(AsyncValue, flfeatures.BoolValue):
    pass


class AsyncIntValue(AsyncNumberValue, flfeatures.IntValue):
    pass


class AsyncRegisterValue(AsyncNumberValue, flfeatures.RegisterValue):
    async def aget_as_number(self):
        value = await self._interface.aquery(self.command_get)
        self._check_value(value)
        return value

    async def aget(self):
        number = await self.aget_as_number()
        return self.compute_dict_from_number(number)

    async def aset(self, value):
        if isinstance(value, dict):
            value = self.compute_number_from_dict(value)

        self._check_value(value)
        await self._interface.awrite(self.command_set + f" {value}")



"""Xiseq file parser (:mod:`pymanip.video.ximea.xiseq`)
=======================================================

.. autoclass:: XiseqFile
   :members:
   :private-members:

"""

from functools import cached_property
from pathlib import Path
import xml.etree.ElementTree as ET


class XiseqFile:
    """Simple parser for Ximea Xiseq file.

    :param filepath: path to file
    :type filepath: :class:`~pathlib.Path` or str

    """

    def __init__(self, filepath):
        """Constructor method
        """
        self.filepath = filepath

    @cached_property
    def tree(self):
        return ET.parse(self.filepath)

    @cached_property
    def metadata_tree(self):
        metadata, = self.tree.findall("imageMetadata")
        return metadata

    @cached_property
    def metadata(self):
        return {child.tag: child.text for child in self.metadata_tree}

    def files(self):
        """Iterate through files in sequence. Yields a dictionnary with keys
        ``timestamp``, ``frame`` and ``filename``.
        """
        for f in self.tree.findall("file"):
            yield {
                "timestamp": f.get("timestamp"),
                "frame": f.get("frame"),
                "filename": f.text,
            }



"""Shortcut to FluidLab interface classes (:mod:`pymanip.interfaces`)
=====================================================================

This module auto-imports all FluidLab interface classes.

"""

from .fluidlab_interfaces import *



"""Andor module (:mod:`pymanip.video.andor`)
============================================

This module is a shortcut for the :class:`pymanip.video.andor.camera.Andor_Camera`
class.

"""

from .camera import Andor_Camera



"""

pymanip CLI interface

"""

from pathlib import Path
from argparse import ArgumentParser
from pymanip.util.session import manip_info, check_hdf, rebuild_from_dat
from pymanip.util.gpib import scanGpib
from pymanip.util.video import (
    preview_pco,
    preview_avt,
    preview_andor,
    preview_ids,
    preview_ximea,
    preview_photometrics,
)

try:
    from pymanip.util.oscillo import Oscillo
    from pymanip.util.channel_selector import ChannelSelector
except ModuleNotFoundError:
    has_oscillo = False
has_video = True

try:
    from pymanip.daq import DAQmx

    has_daq = True
except (ModuleNotFoundError, NotImplementedError):
    has_daq = False

# Create top-level parser
parser = ArgumentParser(description=__doc__, prog="pymanip")
# parser.add_argument('command',
#                    help="pymanip command",
#                    metavar="command")
subparsers = parser.add_subparsers(
    title="command", help="pymanip command", dest="command"
)

# Create parser for the "info" command
parser_info = subparsers.add_parser(
    "info", help="shows the content of a saved pymanip session"
)
parser_info.add_argument(
    "sessionName", help="name of the saved session to inspect", metavar="session_name"
)
parser_info.add_argument(
    "-q", "--quiet", action="store_true", help="do not list content."
)
parser_info.add_argument(
    "-l",
    "--line",
    help="print specified line of logged data.",
    type=int,
    metavar="line",
)
parser_info.add_argument(
    "-p", "--plot", help="plot the specified variable.", metavar="varname"
)

# Create parser for "list_instruments"
parser_list_inst = subparsers.add_parser(
    "list_instruments", help="List supported instruments"
)

# Create parser for "list_daq"
parser_daq = subparsers.add_parser("list_daq", help="List available acquisition cards")

# Create parser for "check_hdf"
parser_hdf = subparsers.add_parser(
    "check_hdf", help="checks dat and hdf files are identical"
)
parser_hdf.add_argument(
    "sessionName",
    help="Name of the pymanip acquisition to inspect",
    metavar="session_name",
)
parser_hdf.add_argument(
    "-p", "--plot", help="Plot the specified variable", metavar="varname"
)

# Create parser for "rebuild_hdf"
parser_rebuild = subparsers.add_parser(
    "rebuild_hdf", help="Rebuilds a pymanip HDF5 file from the ASCII dat file"
)
parser_rebuild.add_argument("input_file", help="Input ASCII file")
parser_rebuild.add_argument("output_name", help="Output MI session name")

# Create parser for "scan_gpib"
parser_gpib = subparsers.add_parser(
    "scan_gpib",
    help="Scans for connected instruments on the specified GPIB board (linux-gpib only)",
)
parser_gpib.add_argument(
    "boardNumber",
    help="GPIB board to scan for connected instruments",
    metavar="board_number",
    type=int,
    default=0,
    nargs="?",
)

# Create parser for "oscillo"
parser_oscillo = subparsers.add_parser(
    "oscillo",
    help="Use NI-DAQmx and NI-Scope cards as oscilloscope and signal analyser",
)
parser_oscillo.add_argument(
    "channel",
    help="DAQmx channel names",
    metavar="channel_name",
    nargs="*",
    default=None,
)
parser_oscillo.add_argument(
    "-s", "--sampling", help="Sampling frequency", metavar="sampling_freq", default=5e3
)
parser_oscillo.add_argument(
    "-r", "--range", help="Channel volt range", metavar="volt_range", default=10.0
)
parser_oscillo.add_argument(
    "-t", "--trigger", help="Trigger level", metavar="level", default=None
)
parser_oscillo.add_argument(
    "-T", "--trigsource", help="Trigger source index", metavar="0", default=0
)
parser_oscillo.add_argument(
    "-b",
    "--backend",
    help="Choose daqmx or scope backend",
    metavar="daqmx",
    default="daqmx",
)
parser_oscillo.add_argument(
    "-p", "--serialport", help="Arduino Serial port", metavar="port", default=None
)

# Create parser for "video"
parser_video = subparsers.add_parser(
    "video", help="Display video preview for specified camera"
)
parser_video.add_argument(
    "camera_type", help="Camera type: PCO, AVT, DC1394, IDS", metavar="camera_type"
)
parser_video.add_argument(
    "-l", "--list", help="List available cameras", action="store_true"
)
parser_video.add_argument(
    "-i",
    "--interface",
    help="Specify interface",
    metavar="interface",
    default="",
    type=str,
)
parser_video.add_argument(
    "-b",
    "--board",
    help="Camera board address",
    metavar="board",
    default=0,
    type=int,
    nargs="+",
)
parser_video.add_argument(
    "-t",
    "--toolkit",
    help="Graphical toolkit to use: cv or qt",
    metavar="toolkit",
    default="qt",
    type=str,
    nargs=1,
)
parser_video.add_argument(
    "-s",
    "--slice",
    help="Slice image x0, x1, y0, y1 in pixels",
    metavar="slice",
    default=[],
    type=int,
    nargs=4,
)
parser_video.add_argument(
    "-z", "--zoom", help="Zoom factor", metavar="zoom", default=0.5, type=float, nargs=1
)
parser_video.add_argument(
    "-T",
    "--Trigger",
    help="Trigger mode",
    metavar="trigger",
    default=-1,
    type=int,
    nargs=1,
)

parser_video.add_argument(
    "-w",
    "--whitebalance",
    help="Enable auto white balance (for color cameras)",
    action="store_true",
)

parser_video.add_argument(
    "-e",
    "--exposure",
    help="Exposure time (ms)",
    metavar="exposure_ms",
    default=20,
    type=float,
    nargs=1,
)
parser_video.add_argument(
    "-d",
    "--bitdepth",
    help="Bit depth",
    metavar="bitdepth",
    default=12,
    type=int,
    nargs=1,
)
parser_video.add_argument(
    "-f",
    "--framerate",
    help="Acquisition framerate in Herz",
    metavar="framerate",
    default=10.0,
    type=float,
    nargs=1,
)
parser_video.add_argument(
    "-r",
    "--rotate",
    help="Rotate image",
    metavar="angle",
    default=0.0,
    type=float,
    nargs=1,
)

parser_video.add_argument(
    "-R",
    "--ROI",
    help="Set Region of Interest xmin, ymin, xmax, ymax",
    metavar="roi",
    default=None,
    type=int,
    nargs=4,
)

if __name__ == "__main__":
    args = parser.parse_args()

    if args.command == "info":
        manip_info(args.sessionName, args.quiet, args.line, args.plot)
    elif args.command == "list_instruments":
        import pymanip

        pymanip.pymanip_import_verbose = True
        import pymanip.instruments
    elif args.command == "check_hdf":
        check_hdf(args.sessionName, args.plot)
    elif args.command == "list_daq":
        if not has_daq:
            print("DAQmx is not available")
        else:
            DAQmx.print_connected_devices()
    elif args.command == "rebuild_hdf":
        rebuild_from_dat(Path(args.input_file), args.output_name)
    elif args.command == "scan_gpib":
        scanGpib(int(args.boardNumber))
    elif args.command == "oscillo":
        if args.trigger is not None:
            trigger = float(args.trigger)
        else:
            trigger = None
        if args.backend is not None:
            backend = args.backend
        if args.channel:
            channel = args.channel
        else:
            if backend == "arduino":
                channel = [int(input("Arduino Analog input pin? "))]
            else:
                chansel = ChannelSelector()
                backend, channel = chansel.gui_select()
        if args.serialport:
            serialport = args.serialport
        elif backend == "arduino":
            serialport = input("Serial port (COM3, /dev/ttyS3, ...)? ")
        if args.sampling:
            sampling = float(args.sampling)
        else:
            sampling = 5e3
        if args.range:
            range_ = float(args.range)
        else:
            range_ = 5.0 if backend == "arduino" else 10.0
        if backend == "arduino":
            backend_args = [serialport]
        else:
            backend_args = None
        oscillo = Oscillo(
            channel,
            sampling,
            5.0 if backend == "arduino" else range_,
            trigger,
            int(args.trigsource),
            backend=backend,
            backend_args=backend_args,
            N=128 if backend == "arduino" else 1024,
        )
        oscillo.run()
    elif args.command == "video":
        if not has_video:
            print("Video libraries not found")
        if isinstance(args.toolkit, list):
            tk = args.toolkit[0]
        else:
            tk = args.toolkit
        if len(args.slice) < 4:
            slice = None
        else:
            slice = args.slice
        if isinstance(args.zoom, list):
            zoom = args.zoom[0]
        else:
            zoom = args.zoom
        if isinstance(args.Trigger, list):
            Trigger = args.Trigger[0]
        else:
            Trigger = args.Trigger
        if Trigger == -1:
            Trigger = None
        if isinstance(args.exposure, list):
            exposure_ms = args.exposure[0]
        else:
            exposure_ms = args.exposure
        if isinstance(args.board, list):
            board = int(args.board[0])
        else:
            board = int(args.board)
        if isinstance(args.bitdepth, list):
            bitdepth = int(args.bitdepth[0])
        else:
            bitdepth = int(args.bitdepth)
        if isinstance(args.framerate, list):
            framerate = float(args.framerate[0])
        else:
            framerate = float(args.framerate)
        if isinstance(args.rotate, list):
            rotate = float(args.rotate[0])
        else:
            rotate = float(args.rotate)
        if args.ROI is not None and len(args.ROI) != 4:
            raise ValueError("ROI must be 4 integers xmin, ymin, xmax, ymax")
        if args.camera_type.upper() == "PCO":
            if args.list:
                from pymanip.video.pco import print_available_pco_cameras

                print_available_pco_cameras()
            else:
                interface = str(args.interface)
                if not interface:
                    interface = "all"
                preview_pco(
                    interface,
                    board,
                    tk,
                    slice,
                    zoom,
                    Trigger,
                    exposure_ms,
                    rotate=rotate,
                    roi=args.ROI,
                )
        elif args.camera_type.upper() == "AVT":
            if args.list:
                print("Listing cameras not implemented for AVT")
            else:
                preview_avt(
                    board,
                    tk,
                    slice,
                    zoom,
                    Trigger,
                    exposure_ms,
                    rotate=rotate,
                    roi=args.ROI,
                )
        elif args.camera_type.upper() == "ANDOR":
            if args.list:
                print("Listing cameras not implemented for Andor")
            else:
                preview_andor(
                    board,
                    tk,
                    slice,
                    zoom,
                    Trigger,
                    exposure_ms,
                    bitdepth,
                    framerate,
                    rotate=rotate,
                )
        elif args.camera_type.upper() == "IDS":
            if args.list:
                print("Listing cameras not implemented for IDS")
            else:
                preview_ids(
                    board,
                    tk,
                    slice,
                    zoom,
                    Trigger,
                    exposure_ms,
                    bitdepth,
                    framerate,
                    rotate=rotate,
                )
        elif args.camera_type.upper() == "XIMEA":
            if args.list:
                print("Listing Ximea camera not implemented.")
            else:
                print("white_balance =", args.whitebalance)
                if board == 0:
                    board = None
                else:
                    raise NotImplementedError("open via SN")
                preview_ximea(
                    board,
                    tk,
                    slice,
                    zoom,
                    Trigger,
                    exposure_ms,
                    rotate=rotate,
                    white_balance=args.whitebalance,
                    roi=args.ROI,
                )
        elif args.camera_type.upper() in ("PHOTOMETRICS", "TELEDYNE", "KINETIX"):
            if args.list:
                from pyvcam.camera import Camera as PVCamera

                print(PVCamera.get_available_camera_names())
            else:
                preview_photometrics(
                    board,
                    tk,
                    slice,
                    zoom,
                    Trigger,
                    exposure_ms,
                    bitdepth,
                    rotate=rotate,
                    roi=args.ROI,
                )
        else:
            print("Unknown camera type: ", args.camera_type)
    else:
        print("Unknown command `{:}'.")



"""

Async Driver extension for the Newport Model XPS-RL Motion controller

"""

import asyncio
import fluidlab.instruments.motor_controller.newport_xps_rl as fl_np

__all__ = ["AsyncNewportXpsRL"]


class AsyncNewportXpsRL(fl_np.NewportXpsRL):
    async def aopen(self):
        self.reader, self.writer = await asyncio.open_connection(
            self.ip_address, self.port
        )

    async def aclose(self):
        self.writer.close()
        await self.writer.wait_closed()

    async def __aenter__(self):
        await self.aopen()
        return self

    async def __aexit__(self, type_, value, cb):
        await self.aclose()

    async def aquery(self, command_str):
        if isinstance(command_str, str):
            command_str = command_str.encode("ascii")
        self.writer.write(command_str)
        await self.writer.drain()
        chunks = []
        while True:
            chunk = await self.reader.read(1024)
            if len(chunk) > 0:
                chunks.append(chunk)
            if len(chunks) < 1024:
                break
        return self._parse_chunks(chunks)

    async def FirmwareVersionGet(self):
        status, response = await self.aquery("FirmwareVersionGet(char *)")
        if status != 0:
            raise fl_np.NewportXpsRLError(status, response)
        return response

    async def ControllerStatusGet(self):
        status, response = await self.aquery("ControllerStatusGet(int *)")
        if status != 0:
            raise fl_np.NewportXpsRLError(status, response)
        return fl_np.NewportXpsRLControllerStatus(int(response))

    async def Login(self, username, password):
        status, response = await self.aquery(f"Login({username}, {password}")
        if status != 0:
            raise fl_np.NewportXpsRLError(status, response)

    async def GroupPositionSetpointGet(self, groupname="Group1.Pos"):
        """
        The SetpointPosition is the profiler position. This is the position where the
        positioner should be according to the ideal theoretical motion profile.
        """

        status, response = await self.aquery(
            f"GroupPositionSetpointGet({groupname}, double *)"
        )
        if status != 0:
            raise fl_np.NewportXpsRLError(status, response)
        return float(response)

    async def GroupPositionCurrentGet(self, groupname="Group1.Pos"):
        """
        The CurrentPosition is the encoder position of the stage after mapping corrections are
        applied. This is the actual position of the positioner at this moment of the query.
        """

        status, response = await self.aquery(
            f"GroupPositionCurrentGet({groupname}, double *)"
        )
        if status != 0:
            raise fl_np.NewportXpsRLError(status, response)
        return float(response)

    async def GroupPositionTargetGet(self, groupname="Group1.Pos"):
        """
        The TargetPosition is the final target position commanded by the user.
        """

        status, response = await self.aquery(
            f"GroupPositionTargetGet({groupname}, double *)"
        )
        if status != 0:
            raise fl_np.NewportXpsRLError(status, response)
        return float(response)

    async def GroupMoveAbsolute(self, groupname="Group1.Pos", target=250.0):
        """
        Initiates an absolute move for a positioner or a group.
        """

        status, response = await self.aquery(
            "GroupMoveAbsolute({:}, {:.2f})".format(groupname, float(target))
        )
        if status != 0:
            raise fl_np.NewportXpsRLError(status, response)



"""Asynchronous TTI CPX400dp extension (:mod:`pymanip.aioinstruments.tti_cpx400dp`)
===================================================================================

.. autoclass:: AsyncTtiCp400dp
   :members:
   :private-members:

"""

import fluidlab.instruments.powersupply.tti_cpx400dp as fl_tti
from pymanip.aioinstruments.aiofeatures import AsyncFloatValue, AsyncBoolValue
from pymanip.aioinstruments.aioiec60488 import AsyncIEC60488


class AsyncTtiCpx400dpUnitValue(AsyncFloatValue, fl_tti.TtiCpx400dpUnitValue):
    pass


class AsyncTtiCpx400dp(AsyncIEC60488, fl_tti.TtiCpx400dp):
    pass


afeatures = [
    AsyncTtiCpx400dpUnitValue(
        "vdc",
        doc="Get actual voltage/Set voltage setpoint on specified channel",
        command_set="V{channel:d} {value}",
        command_get="V{channel:d}O?",
        channel_argument=True,
        check_instrument_value=False,
    ),
    AsyncTtiCpx400dpUnitValue(
        "idc",
        doc="Get actual current/Set current setpoint on specified channel",
        command_set="I{channel:d} {value}",
        command_get="I{channel:d}O?",
        channel_argument=True,
        check_instrument_value=False,
    ),
    AsyncBoolValue(
        "onoff",
        doc="Toogle output ON/OFF for specified channel",
        command_set="OP{channel:d} {value}",
        command_get="OP{channel:d}?",
        channel_argument=True,
        check_instrument_value=False,
        true_string="1",
        false_string="0",
    ),
    AsyncBoolValue(
        "onoffall",
        doc="Toogle output ON/OFF for both channels simultaneously",
        command_set="OPALL ",
        channel_argument=False,
        check_instrument_value=False,
        true_string="1",
        false_string="0",
    ),
]

AsyncTtiCpx400dp._build_class_with_features(afeatures)



"""Concrete implementation with arduino ino file (:mod:`pymanip.aiodaq.arduino`)
================================================================================

This is just a convenient placeholder, since all methods are defined in
:mod:`pymanip.aioinstruments.arduino_gbf`.

"""

from pymanip.aioinstruments.arduino_gbf import AsyncArduino as ArduinoSystem



"""

This module defines utility function to interact with pymanip sessions

"""

import sys
import time
import os

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.dates import AutoDateFormatter, AutoDateLocator, epoch2num
from fluiddyn.util.terminal_colors import cprint

try:
    import pandas as pd

    has_panda = True
except ModuleNotFoundError:
    has_panda = False
from pymanip.collection import Manip
from pymanip import Session, SavedSession
from pymanip.asyncsession import SavedAsyncSession
from pymanip.mytime import dateformat


def manip_info(sessionName, quiet, line_to_print, var_to_plot):
    """
    This function prints information about a session, and optionally
    plot specified variables.

    It can be accessed from the CLI tool ManipInfo
    """

    if os.path.exists(sessionName + ".db"):
        SavedAsyncSession(sessionName).print_description()
        return

    if sessionName.endswith(".hdf5"):
        N = len(sessionName)
        sessionName = sessionName[0 : (N - 5)]

    MI = Manip(sessionName).MI
    if line_to_print is not None:
        if line_to_print >= len(MI.log("t")):
            print("Specified line is out of bound.")
            sys.exit(1)
        format_str = "{:>15} | {:>20}"
        print("Printing saved values on line", line_to_print)
        print(format_str.format("Variable", "Value"))
        varlist = ["Time"]
        varlist += MI.log_variable_list()
        print("-" * 38)
        for varname in varlist:
            valtab = MI.log(varname)
            if isinstance(valtab, (float, int)):
                # might occur if only one line
                print(format_str.format(varname, MI.log(varname)))
            else:
                print(format_str.format(varname, MI.log(varname)[line_to_print]))
    elif not quiet:
        MI.describe()
    if var_to_plot is not None:
        if var_to_plot in MI.log_variable_list():
            t = epoch2num(MI.log("t"))
            vardata = MI.log(var_to_plot)
            fig = plt.figure()
            xtick_locator = AutoDateLocator()
            xtick_formatter = AutoDateFormatter(xtick_locator)
            ax = plt.axes()
            ax.xaxis.set_major_locator(xtick_locator)
            ax.xaxis.set_major_formatter(xtick_formatter)
            ax.plot(t, vardata, "o-")
            plt.setp(ax.xaxis.get_majorticklabels(), rotation=70)
            fig.subplots_adjust(bottom=0.2)
            plt.ylabel(var_to_plot)
            plt.title(sessionName)
            plt.show()
        else:
            print("Variable", var_to_plot, "does not exist!")
            sys.exit(1)


def check_hdf(acqName, variable_to_plot):
    """
    This functions checks that the .dat file and the .hdf5 file of
    a pymanip session holds the same data.
    """

    if not has_panda:
        print("Pandas must be installed.")
        sys.exit(-1)

    # Lecture HDF5
    MI = SavedSession(acqName)

    # Lecture DAT
    datfile = acqName + ".dat"
    print("Loading saved session from file", datfile)
    data = pd.read_csv(datfile, sep=" ")
    start_t = data["Time"].values[0]
    end_t = data["Time"].values[-1]
    start_string = time.strftime(dateformat, time.localtime(start_t))
    end_string = time.strftime(dateformat, time.localtime(end_t))
    cprint.blue("*** Start date: " + start_string)
    cprint.blue("***  End date: " + end_string)

    varlist = ["Time"]
    varlist += MI.log_variable_list()

    # First line
    print("Printing first line.")
    format_str = "{:>15} | {:>20} | {:>20}"
    print(format_str.format("", "HDF", "DAT"))
    print("-" * 80)
    for varname in varlist:
        data_hdf = MI.log(varname)[0]
        data_dat = data[varname].values[0]
        print(format_str.format(varname, data_hdf, data_dat))

    # Comparaison
    length_warning_done = False
    for varname in varlist:
        var_hdf = MI.log(varname)
        var_dat = data[varname].values
        if len(var_hdf) != len(var_dat):
            common_length = np.min([len(var_hdf), len(var_dat)])
            if not length_warning_done:
                print("Number of data points differs.")
                print("HDF:", len(var_hdf))
                print("DAT:", len(var_dat))
                length_warning_done = True
                print("Comparing the first", common_length, " values")
            var_hdf = var_hdf[0:common_length]
            var_dat = var_dat[0:common_length]
        reldiff = np.abs(var_hdf - var_dat) / (var_hdf + var_dat)
        bb = reldiff > 0.01
        if bb.any():
            print("Data differs for variable " + varname)
            index = np.min(bb.argmax())
            print("On line", index, ": HDF=", var_hdf[index], "DAT=", var_dat[index])
        else:
            print("Identical data for variable " + varname)

    # Plot
    if variable_to_plot is not None:
        plt.figure(1)
        t = MI.log("t")
        var = MI.log(variable_to_plot)
        plt.plot(t - t[0], var)
        plt.title("HDF")
        plt.xlabel("t")
        plt.ylabel(variable_to_plot)

        plt.figure(2)
        t = data["Time"].values
        var = data[variable_to_plot].values
        plt.plot(t - t[0], var)
        plt.title("DAT")
        plt.xlabel("t")
        plt.ylabel(variable_to_plot)

        plt.show()


def rebuild_from_dat(inputDatfile, outputSessionName):
    """
    Rebuilds a pymanip HDF5 file from the ASCII dat file.
    """

    if not has_panda:
        print("Pandas is not available.")
    else:
        with inputDatfile.open() as in_f:
            data = pd.read_csv(in_f, sep=" ")
            liste_var = list(data.keys())
            liste_var.remove("Time")
            MI = Session(outputSessionName, liste_var)
            for line in data.iterrows():
                MI.log_addline(timestamp=line[1].Time, dict_caller=dict(line[1]))
            MI.Stop()



"""Ximea module (:mod:`pymanip.video.ximea`)
========================================

This module is a shortcut for the :class:`pymanip.video.ximea.camera.Ximea_Camera`
class.

"""

from .camera import Ximea_Camera


import os
import asyncio
import pytest
import numpy as np
from pymanip.asyncsession import AsyncSession


def test_logged_variables():
    async def task(sesn):
        sesn.add_entry(a=sesn.a)
        sesn.a = sesn.a + 1
        await asyncio.sleep(0.001)
        if sesn.a == 4:
            sesn.ask_exit()

    async def main(sesn):
        await sesn.monitor(task, server_port=None)

    with AsyncSession() as sesn:
        sesn.a = 0
        asyncio.run(main(sesn))
        # Test that sesn.logged_variable('varname') is equivalent to sesn['varname']
        # and returns timestamps and values
        ta, va = sesn.logged_variable("a")
        tta, vva = sesn["a"]
        assert (ta == tta).all()
        assert (va == vva).all()
        assert (va == [0, 1, 2, 3]).all()
        # Test the other methods
        assert sesn.logged_variables() == set("a")
        data = sesn.logged_data()
        dta, vta = data["a"]
        assert (dta == ta).all()
        assert (vta == va).all()
        assert sesn.logged_first_values() == {"a": (ta[0], 0)}
        assert sesn.logged_last_values() == {"a": (ta[-1], 3)}
        ttta, vvva = sesn.logged_data_fromtimestamp("a", sesn.t0)
        assert (ttta == ta).all()
        assert (vvva == va).all()


def test_run_monitor_equivalence():
    async def task(sesn):
        await asyncio.sleep(0.001)
        sesn.toto = 1
        sesn.ask_exit()

    async def main(sesn):
        await sesn.monitor(task, server_port=None)

    with AsyncSession() as sesn:
        sesn.toto = 0
        asyncio.run(main(sesn))
        assert sesn.toto == 1

    with AsyncSession() as sesn:
        sesn.toto = 0
        sesn.run(task, server_port=None)
        assert sesn.toto == 1


def test_parameters(tmpdir):
    async def dummy(sesn):
        sesn.ask_exit()
        await asyncio.sleep(0.001)

    with AsyncSession(os.path.join(tmpdir, "test_asyncsession")) as sesn:
        params = {"c": 3e8, "pi": 3.14}
        sesn.save_parameter(params, a=1, b=2)
        sesn.save_parameter(d=10)
        sesn.run(dummy, server_port=None)

    sesn = None

    with AsyncSession(os.path.join(tmpdir, "test_asyncsession")) as sesn:
        a = sesn.parameter("a")
        params = sesn.parameters()

        assert sesn.has_parameter("pi")

        b = params["b"]
        c = params["c"]
        pi = params["pi"]
        d = params["d"]

        assert a == 1
        assert b == 2
        assert c == 3e8
        assert pi == 3.14
        assert d == 10


def test_datasets():
    async def task(sesn):
        sesn.add_dataset(a=[1, 2, 3, 4])
        sesn.add_dataset(c=[1, 2, 3])
        await asyncio.sleep(0.001)
        sesn.add_dataset(b=np.array([5, 6, 7, 8]))
        sesn.add_dataset(c=[4, 5, 6])
        sesn.ask_exit()

    with AsyncSession() as sesn:
        sesn.run(task, server_port=None)
        assert sesn.dataset("a") == [1, 2, 3, 4]
        assert (sesn.dataset("b") == (5, 6, 7, 8)).all()
        for ts, d_a in sesn.datasets("a"):
            assert d_a == [1, 2, 3, 4]
        assert sesn.dataset_names() == {"a", "b", "c"}
        t_a, d_a = sesn.dataset_last_data("a")
        assert d_a == [1, 2, 3, 4]
        t_a = sesn.dataset_times("a")[0]
        t_b = sesn.dataset_times("b")[0]
        assert t_b - t_a >= 0.001
        t_last, data_last = sesn.dataset_last_data("c")
        assert data_last == [4, 5, 6]
